""" HalfEdge: The intermediate level datastructure of the dcel """
##-- imports
from __future__ import annotations

import logging as root_logger
from dataclasses import InitVar, dataclass, field
from itertools import cycle, islice
from math import atan2, degrees, pi
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

import numpy as np

from ..constants import D_EPSILON, EDGE, END, EPSILON, START, TOLERANCE, TWOPI
from ..drawing import clear_canvas, draw_circle, draw_text
from ..umath import (bbox_to_lines, extend_line, get_distance_raw,
                     get_midpoint, get_ranges, get_unit_vector, in_circle,
                     intersect, is_point_on_line, rotate_point,
                     sample_along_lines)
from .constants import EDGE_FOLLOW_GUARD, EdgeE, EditE
from .drawable import Drawable
from .line import Line
from .vertex import Vertex

##-- end imports

logging = root_logger.getLogger(__name__)

PI     = pi
TWOPI  = 2 * PI
HALFPI = PI * 0.5
QPI    = PI * 0.5

@dataclass
class HalfEdge(Drawable):
    """
    A Canonical Half-Edge. Has an origin point, and a twin
    half-edge for its end point,
    Auto-maintains counter-clockwise vertex order with it's twin.
    Two HalfEdges make an Edge
    """

    origin             : Vertex         = field()
    twin               : 'HalfEdge'     = field()
    dcel               : 'DCEL'         = field()
    data               : Dict[Any, Any] = field(default_factory=dict)
    index              : int            = field(default=None)
    length_sq          : float          = field(default=-1)
    face               : Face           = field(default=None)
    next               : 'HalfEdge'     = field(default=None)
    prev               : 'HalfEdge'     = field(default=None)

    marked_for_cleanup : bool           = field(init=False, default=False)
    constrained        : bool           = field(init=False, default=False)
    drawn              : bool           = field(init=False, default=False)
    fixed              : bool           = field(init=False, default=False)

    nextIndex = 0

    def __post_init__(self):
        assert(origin is None or isinstance(origin, Vertex))
        assert(twin is None or isinstance(twin, HalfEdge))

        if index is None:
            logging.debug("Creating Edge {}".format(HalfEdge.nextIndex))
            self.index = HalfEdge.nextIndex
            HalfEdge.nextIndex += 1
        else:
            assert(isinstance(index, int))
            logging.debug("Re-creating Edge: {}".format(index))
            self.index = index
            if self.index >= HalfEdge.nextIndex:
                HalfEdge.nextIndex = self.index + 1

        #register the halfedge with the vertex
        if origin is not None:
            self.origin.register_half_edge(self)

        if self.dcel is not None and self not in self.dcel.half_edges:
            self.dcel.half_edges.add(self)

    def eq_verts(self, other):
        """ Test whether two halfedges share vertices """
        assert(isinstance(other, HalfEdge))
        s_verts = self.get_vertices()
        o_verts = other.get_vertices()
        return all([s == o for s, o in zip(s_verts, o_verts)])

    def copy(self):
        """ Copy the halfedge pair. sub-copies the vertexs too """
        assert(self.origin is not None)
        assert(self.twin is not None)
        #copy the vertices
        v1 = self.origin.copy()
        v2 = self.twin.origin.copy()
        #create the halfedge
        e = self.dcel.new_edge(v1, v2)
        #update next/prev?

        #copy data
        e.data.update(self.data)
        return e


    #------------------------------
    # def export
    #------------------------------
    def _export(self):
        """ Export identifiers instead of objects to allow reconstruction """
        logging.debug("Exporting Edge: {}".format(self.index))
        origin = self.origin
        if origin is not None:
            origin = origin.index
        twin = self.twin
        if twin is not None:
            twin = twin.index
        face = self.face
        if face is not None:
            face = face.index
        next_he = None
        prev_he = None
        if self.next is not None:
            next_he = self.next.index
        if self.prev is not None:
            prev_he = self.prev.index
        enum_data = {a.name:b for a, b in self.data.items() if a in EdgeE}
        non_enum_data = {a:b for a, b in self.data.items() if a not in EdgeE}


        return {
            'i' : self.index,
            'origin' : origin,
            'twin' : twin,
            'face' : face,
            'next' : next_he,
            'prev' : prev_he,
            "enum_data" : enum_data,
            "non_enum_data": non_enum_data
        }


    #------------------------------
    # def Human Readable Representations
    #------------------------------

    def __repr__(self):
        return "<HalfEdge {}: {} - {}>".format(self.index, self.origin, self.twin.origin)

    def __str__(self):
        origin     = "n/a"
        twin       = "n/a"
        n          = "n/a"
        p          = "n/a"
        f          = "n/a"

        if self.origin is not None:
            origin = self.origin.index
        if self.twin is not None:
            twin   = self.twin.index
        if self.next is not None:
            n      = self.next.index
        if self.prev is not None:
            p      = self.prev.index
        if self.face is not None:
            f      = self.face.index

        coords     = [str(x) for x in self.get_vertices()]

        data       = (self.index, f, origin, twin, p, n, coords)
        return "(HE: {}, f: {}, O: {}, T: {}, P: {}, N: {}, XY: {})".format(*data)

    def draw(self, ctx, data_override=None, clear=False, text=False, width=None):
        #pylint: disable=too-many-locals
        logging.debug("Drawing Edge: {} | {}".format(self.index, self.twin.index))
        if clear:
            clear_canvas(ctx)
        data = self.data.copy()
        if data_override is not None:
            assert(isinstance(data_override, dict))
            data.update(data_override)

        #defaults
        colour             = EDGE
        start_end_points   = False
        start_col          = START
        end_col            = END
        start_rad          = width
        end_rad            = width
        write_text         = "HE:{}.{}".format(self.index, self.twin.index)
        bezier             = False
        bezier_simp        = False
        sample_description = None

        #retrieve custom values
        if EdgeE.WIDTH in data:
            width              = data[EdgeE.WIDTH]
            start_rad          = width
            end_rad            = width
        if EdgeE.STROKE in data:
            colour             = data[EdgeE.STROKE]
        if EdgeE.START in data and isinstance(data[EdgeE.START], (list, np.ndarray)):
            start_col          = data[EdgeE.START]
        if EdgeE.END in data and isinstance(data[EdgeE.END], (list, np.ndarray)):
            end_col            = data[EdgeE.END]
        if EdgeE.START in data and EdgeE.END in data:
            start_end_points   = True
        if EdgeE.STARTRAD in data:
            start_rad          = data[EdgeE.STARTRAD]
        if EdgeE.ENDRAD in data:
            end_rad            = data[EdgeE.ENDRAD]
        if EdgeE.TEXT in data:
            if isinstance(data[EdgeE.TEXT], str):
                write_text     = data[EdgeE.TEXT]
            elif not data[EdgeE.TEXT]:
                write_text     = None
        if EdgeE.BEZIER in data:
            bezier             = data[EdgeE.BEZIER]
            assert(isinstance(bezier, list))
        if EdgeE.BEZIER_SIMPLIFY in data:
            bezier_simp        = True
        if EdgeE.SAMPLE in data:
            sample_description = data[EdgeE.SAMPLE]

        #Get Start and end points
        v1, v2 = self.get_vertices()
        if v1 is None or v2 is None:
            #early exit if line is not completed
            return
        centre = get_midpoint(v1.to_array(), v2.to_array())
        logging.debug("Drawing HalfEdge {} : {}, {} - {}, {}".format(self.index,
                                                                     v1.loc[0],
                                                                     v1.loc[1],
                                                                     v2.loc[0],
                                                                     v2.loc[1]))
        if sample_description is not None:
            #draw as a sampled line
            sample_description(ctx, self)

        if EdgeE.NULL in data:
            return

        ctx.set_line_width(width)
        ctx.set_source_rgba(*colour)

        #draw as a line/curve
        #todo: allow beziers to be simplified to straight lines
        if bool(bezier):
            logging.debug("Drawing Bezier: {}".format(bezier))
            ctx.new_path()
            for b in bezier:
                ctx.move_to(*b[0])
                if bezier_simp:
                    ctx.line_to(*b[-1])
                    continue
                if len(b) == 3:
                    ctx.curve_to(*b[0], *b[1], *b[2])
                else:
                    assert(len(b) == 4)
                    ctx.curve_to(*b[1], *b[2], *b[3])

        else:
            logging.debug("Drawing Straight Line")
            ctx.move_to(*v1.loc)
            ctx.line_to(*v2.loc)

        ctx.stroke()

        if start_end_points:
            ctx.set_source_rgba(*start_col)
            draw_circle(ctx, *v1.loc, start_rad)
            ctx.set_source_rgba(*end_col)
            draw_circle(ctx, *v2.loc, end_rad)

        if text and write_text is not None:
            draw_text(ctx, *centre, write_text)


    #------------------------------
    # def Math
    #------------------------------

    def cross(self):
        """ Cross product of the halfedge """
        assert(self.origin is not None)
        assert(self.twin is not None)
        assert(self.twin.origin is not None)
        a = self.origin.to_array()
        b = self.twin.origin.to_array()
        return np.cross(a, b)

    def get_length_sq(self, force=False):
        """ Gets the calculated length, or calculate it. returns as a np.ndarray"""
        if not force and self.length_sq is not -1:
            return self.length_sq
        #otherwise calculate
        as_array = self.to_array()
        self.length_sq = get_distance_raw(as_array[0], as_array[1])
        return self.length_sq


    #------------------------------
    # def Modifiers
    #------------------------------

    def split(self, loc, copy_data=True, face_update=True):
        """ Take an s -> e, and make it now two edges s -> (x, y) -> e
        returns (firstHalf, new_point, secondHalf)"""
        assert(isinstance(loc, (np.ndarray, Vertex)))
        start = self.origin
        end = self.twin.origin
        if isinstance(loc, Vertex):
            new_point = loc
        else:
            new_point = self.dcel.new_vertex(loc)
        if copy_data:
            new_point.data.update(start.data)
        new_edge = self.dcel.new_edge(new_point, end)
        if copy_data:
            new_edge.data.update(self.data)
        #update the twin
        self.twin.origin = new_point
        #update registrations:
        end.unregister_half_edge(self)
        new_point.register_half_edge(self)
        new_point.register_half_edge(self.twin)
        end.unregister_half_edge(self.twin)
        end.register_half_edge(new_edge.twin)
        #recalculate length
        self.get_length_sq(force=True)
        self.twin.get_length_sq(force=True)
        #insert into next/prev ordering
        new_edge.add_next(self.next, force=True)
        new_edge.twin.add_prev(self.twin.prev, force=True)
        self.add_next(new_edge, force=True)
        new_edge.twin.add_next(self.twin, force=True)
        #update faces
        if face_update and self.face is not None:
            self.face.add_edge(new_edge)
        if face_update and self.twin.face is not None:
            self.twin.face.add_edge(new_edge.twin)
        return (new_point, new_edge)

    def split_by_ratio(self, r=0.5, face_update=True):
        """ Split an edge by a ratio of 0.0 - 1.0 : start - end.
        defaults to 0.5, the middle """
        point = sample_along_lines((self.to_array().flatten()), 1, override=np.array([r]))
        return self.split(point[0], face_update=face_update)

    def translate(self, direction, d=1, absolute=False, candidates=None, force=False):
        """ Move the edge by a vector and distance, or to an absoluteolute location """
        assert(isinstance(direction, np.ndarray))
        if not absolute:
            target = self.to_array() + (direction * d)
        else:
            assert(direction.shape == (2, 2))
            target = direction

        if not force and self.has_constraints(candidates):
            return (self.dcel.create_edge(target[0],
                                         target[1],
                                         edata=self.data,
                                         vdata=self.origin.data), EditE.NEW)
        else:
            _, edit1 = self.origin.translate(target[0], absolute=True, force=True)
            _, edit2 = self.twin.origin.translate(target[1], absolute=True, force=True)
            assert(edit1 == edit2)
            assert(edit1 == EditE.MODIFIED)
            return (self, EditE.MODIFIED)

    def extend(self, target=None, direction=None, rotate=None, d=1, in_sequence=True):
        """ Extend the line with a new line in the direction of 'target',
        or in the normalized direction 'direction', by distance d.
        if no target or direction is passed in, it extends in the line direction """
        start = self.origin.to_array()
        end = self.twin.origin.to_array()
        new_end = None
        if sum([1 for x in [target, direction, rotate] if x is not None]) > 1:
            raise Exception("HalfEdge.extend: Specify only one of target, direction, rotate")
        if target is not None:
            assert(isinstance(target, np.ndarray))
            assert(len(target) == 2)
            if d is not None:
                new_end = extend_line(end, target, d, from_start=False)
            else:
                new_end = target
        elif direction is not None:
            #use the direction raw
            assert(hasattr(direction, "__len__"))
            assert(len(direction) == 2)
            assert(d is not None)
            new_end = extend_line(end, end + direction, d)
        elif rotate is not None:
            #rotate the vector of the existing line and extend by that
            unit_vector = get_unit_vector(start, end)
            rotated = rotate_point(unit_vector, np.array([0, 0]), rads=rotate)
            new_end = extend_line(end, end + rotated, d)
        else:
            assert(d is not None)
            #get the normalized direction of self.origin -> self.twin.origin
            new_end = extend_line(start, end, from_start=False)
        #Then create a point at (dir * d), create a new edge to it
        new_vert = self.dcel.new_vertex(new_end)

            #todo: twinNext is the next ccw edge for the correct face

        new_edge = self.dcel.new_edge(self.twin.origin, new_vert,
                                      edata=self.data, vdata=self.origin.data)
        new_edge.fix_faces(self)

        return new_edge

    def rotate(self, c=None, r=0, candidates=None, force=False):
        """ return Rotated coordinates as if the edge was rotated around a point by rads """
        assert(isinstance(c, np.ndarray))
        assert(c.shape == (2, ))
        assert(-TWOPI <= r <= TWOPI)
        as_array = self.to_array()
        rotated_coords = rotate_point(as_array, cen=c, rads=r)

        if not force and self.has_constraints(candidates):
            return (self.dcel.create_edge(rotated_coords[0],
                                         rotated_coords[1],
                                         edata=self.data,
                                         vdata=self.origin.data), EditE.NEW)
        else:
            _, edit1 = self.origin.translate(rotated_coords[0], absolute=True, force=True)
            _, edit2 = self.twin.origin.translate(rotated_coords[1], absolute=True, force=True)
            assert(edit1 == edit2)
            assert(edit1 == EditE.MODIFIED)
            return (self, EditE.MODIFIED)

    def constrain_to_circle(self, centre, radius, candidates=None, force=False):
        """ Modify or create a new edge that is constrained to within a circle,
        while also marking the original edge for cleanup if necessary """
        #pylint: disable=too-many-locals
        #todo: handle sequences
        assert(isinstance(centre, np.ndarray))
        assert(centre.shape == (2, ))
        assert(radius >= 0)
        results = self.within_circle(centre, radius)
        logging.debug("HE: within_circle? {}".format(results))
        if all(results):
            #nothing needs to be done
            logging.debug("HE: fine")
            return (self, EditE.MODIFIED)
        if not any(results):
            logging.debug("HE: to remove")
            self.mark_for_cleanup()
            return (self, EditE.MODIFIED)

        closer, further = self.get_closer_and_further(centre)
        as_line = Line.new_line(np.array([closer.to_array(), further.to_array()]))
        intersections = as_line.intersect_with_circle(centre, radius)

        distances = get_distance_raw(further.to_array(), intersections)
        closest = intersections[np.argmin(distances)]
        vert_target = None

        if not force and self.has_constraints(candidates):
            edit_e = EditE.NEW
            vert_target = self.dcel.new_vertex(closest)
            target = self.copy()
            self.mark_for_cleanup()
        else:
            edit_e = EditE.MODIFIED
            target = self

        if further == self.origin:
            if vert_target is not None:
                target.replace_vertex(vert_target)
            else:
                target.origin.loc = closest
        else:
            if vert_target is not None:
                target.twin.replace_vertex(vert_target)
            else:
                target.twin.origin.loc = closest


        return (target, edit_e)

    def constrain_to_bbox(self, bbox, candidates=None, force=False):
        """ force this edge within a bbox, or a new edge to be """
        #pylint: disable=too-many-locals
        if not force and self.has_constraints(candidates):
            edge_prime, _ = self.copy().constrain_to_bbox(bbox, force=True)
            return (edge_prime, EditE.NEW)

        #get intersections with bbox
        intersections = self.intersects_bbox(bbox)
        verts = self.get_vertices()
        vert_coords = self.to_array()

        if self.within(bbox):
            logging.debug("Ignoring halfedge: is within bbox")
        elif self.outside(bbox):
            self.mark_for_cleanup()
        elif not bool(intersections):
            raise Exception("Edge Constraining: Part in and out, with no intersection")
        elif len(intersections) == 1:
            logging.debug("One intersection, moving outside vertex")
            intersect_coords, _ = intersections[0]
            outside_verts = [x for x in verts if not x.within(bbox)]
            assert(len(outside_verts) == 1)
            if outside_verts[0] == self.origin:
                d = self.origin.data
                target = self
            else:
                d = self.twin.origin.data
                target = self.twin
            new_vert = self.dcel.new_vertex(intersect_coords, data=d)
            target.replace_vertex(new_vert)

        elif len(intersections) == 2:
            logging.debug("Two intersections, moving both vertices")
            for i_c, _ in intersections:
                vert_to_move = verts[np.argmin(get_distance_raw(vert_coords, i_c))]
                if vert_to_move == self.origin:
                    d = self.origin.data
                    target = self
                else:
                    d = self.twin.origin.data
                    target = self.twin
                new_vert = self.dcel.new_vertex(intersect_coords, data=d)
                target.replace_vertex(new_vert)


        return (self, EditE.MODIFIED)


    #------------------------------
    # def Comparison
    #------------------------------

    def intersect(self, other_edge):
        """ Intersect two edges mathematically,
        returns intersection point or None """
        assert(isinstance(other_edge, HalfEdge))
        line_segment_1 = self.to_array()
        line_segment_2 = other_edge.to_array()
        return intersect(line_segment_1, line_segment_2)

    def intersects_bbox(self, bbox, tolerance=TOLERANCE):
        """ Return an enum of the edges of a bbox the line intersects
        returns a cuty.constants.IntersectEnum
        returns a list. empty list is no intersections

            bbox is [min_x, min_y, max_x, max_y]
        """
        #calculate intersection points for each of the 4 edges of the bbox,
        #return as tuple of tuples: [( IntersectEnum, np.array(coordinates) )]

        assert(isinstance(bbox, np.ndarray))
        assert(len(bbox) == 4)
        if self.origin is None or self.twin.origin is None:
            raise Exception("Invalid line boundary test ")
        #adjust the bbox by an epsilon? not sure why. TODO: test this
        bbox_lines = bbox_to_lines(bbox)
        self_line_segment = self.to_array()
        start, end = self.to_array()

        logging.debug("Checking edge intersection:\n {}\n {}\n->{}\n----".format(start,
                                                                                 end,
                                                                                 bbox))
        result = []
        #run the 4 intersections
        for (curr_line, enum_value) in bbox_lines:
            intersected = intersect(self_line_segment, curr_line, tolerance=tolerance)
            if intersected is not None:
                result.append((intersected, enum_value))

        assert(len(result) < 3)
        return result

    def point_is_on_line(self, point):
        """ Test to see if a particular x, y coord is on a line """
        assert(isinstance(point, np.ndarray))
        assert(point.shape == (2, ))
        coords = self.to_array()
        return is_point_on_line(point, coords)

    def __call__(self, x=None, y=None):
        """ Pass in a value and calculate the other """
        assert(any([a is not None for a in [x, y]]))
        assert(not all([a is not None for a in [x, y]]))
        the_line = Line.new_line(self.to_array())
        return the_line(x=x, y=y)

    def get_ranges(self):
        """ Get the bbox of the halfedge """
        arr = get_ranges(self.to_array())
        return arr

    @staticmethod
    def compare_edges(center, a, b):
        """ Compare two halfedges against a centre point,
        returning whether a is CCW, equal, or CW from b
        """
        assert(isinstance(center, np.ndarray))
        assert(isinstance(a, HalfEdge))
        assert(isinstance(b, HalfEdge))

        offset_a = a.origin.to_array() - center
        offset_b = b.origin.to_array() - center

        deg_a = (degrees(atan2(offset_a[1], offset_a[0])) + 360) % 360
        deg_b = (degrees(atan2(offset_b[1], offset_b[0])) + 360) % 360

        return deg_a <= deg_b

    def degrees(self, centre):
        """ how many degrees from the center is this halfedge? """
        offset = self.origin.to_array() - centre
        deg = (degrees(atan2(offset[1], offset[0])) + 360) % 360
        return deg

    @staticmethod
    def ccw(a, b, c):
        """ Test for left-turn on three points of a triangle """
        assert(all([isinstance(x, np.ndarray) for x in [a, b, c]]))
        offset_b = b - a
        offset_c = c - a
        crossed = np.cross(offset_b, offset_c)
        return crossed >= 0

    @staticmethod
    def ccw_e(a, b, c):
        """ Test a centre point and two halfedges for ccw ordering """
        assert(isinstance(a, np.ndarray))
        assert(isinstance(b, HalfEdge))
        assert(isinstance(c, HalfEdge))
        first_origin = b.origin.to_array()
        second_origin = c.origin.to_array()
        offset_b = first_origin - a
        offset_c = second_origin - a
        crossed = np.cross(offset_b, offset_c)
        return crossed

    def __lt__(self, other):
        return HalfEdge.compare_edges(self.face.get_centroid(), self, other)

    def he_ccw(self, centre):
        """ Verify the halfedge is ccw ordered """
        assert(isinstance(centre, np.ndarray))
        return HalfEdge.ccw(centre, self.origin.to_array(), self.twin.origin.to_array())

    def is_upper(self):
        """ Is this halfedge the higher of the pair? """
        verts = self.get_vertices()
        return verts[0] < verts[1]

    def is_flat(self):
        """ is this halfedge flat? """
        arr = self.to_array()
        return arr[0, 1] == arr[1, 1]

    def contains_vertex(self, vert, tolerance=D_EPSILON):
        """ is the vertex passed in on the halfedge? """
        assert(isinstance(vert, Vertex))
        verts = self.to_array()
        if vert in verts:
            return True
        l = self.to_array()
        p = vert.to_array()
        return is_point_on_line(p, l)
        # l = Line.new_line(verts)
        # xprime = l(y=vert.loc[1])[0]
        # diff = abs(vert.loc[0]) - abs(xprime)
        # return -(tolerance) <= diff <= tolerance


    #------------------------------
    # def Utilities
    #------------------------------

    @staticmethod
    def avg_direction(edges):
        """ Get the average normalised direction vector of each component of the
        total line segment """
        assert(isinstance(edges, list))
        assert(all([isinstance(x, HalfEdge) for x in edges]))
        all_lines = [Line.new_line(x.to_array()) for x in edges]
        all_directions = np.array([x.direction for x in all_lines])
        direction = all_directions.sum(axis=0) / len(edges)
        return direction

    def vertex_intersections(self, e=EPSILON):
        """ Create a bbox for the total line segment, and intersect check that with the
        dcel quadtree """
        raise Exception("Unimplemented")

    def follow_sequence(self, backwards=False, guard=EDGE_FOLLOW_GUARD):
        """ Follow the .next or .prev chain to completion or loop """
        count = 1
        edges = [self]
        getter = lambda x: x.next
        if backwards:
            getter = lambda x: x.prev
        current = getter(self)
        #todo: possibly use a set and stop on any loop
        while count < guard and current is not None and current is not self:
            edges.append(current)
            current = getter(current)
            count += 1

        return edges


    #------------------------------
    # def Verification
    #------------------------------

    def fix_faces(self, originator):
        """ Infer faces by side on a vertex,
        leftmost means to fix on the right instead """
        extended_from = originator
        all_twins = [x.twin.origin for x in self.origin.half_edges]
        edge_lookup = {x.twin.origin : x.twin for x in self.origin.half_edges}
        assert(extended_from.origin in all_twins)
        ordered = self.dcel.order_vertices(self.origin.loc, all_twins)
        extended_index = ordered.index(extended_from.origin)
        zipped = zip(islice(cycle(ordered), extended_index, len(ordered) + extended_index),
                     islice(cycle(ordered), extended_index+1, len(ordered) + extended_index + 1))


        for a, b in zipped:
            a_edge = edge_lookup[a]
            b_edge = edge_lookup[b]
            a_edge.twin.add_prev(b_edge, force=True)

        if self.prev.face is None:
            new_face = self.dcel.new_face()
            new_face.add_edge(self.prev)
        if originator.twin.face is None:
            orig_twin_face = self.dcel.new_face()
            orig_twin_face.add_edge(originator.twin)


        self.prev.face.add_edge(self)
        f2_sequence = self.twin.follow_sequence()
        if originator.twin in f2_sequence:
            originator.twin.face.add_edge(self.twin)
        else:
            twin_face = self.dcel.new_face()
            for e in f2_sequence:
                self.prev.face.remove_edge(e)
                twin_face.add_edge(e)

    def has_constraints(self, candidate_set=None):
        """ Tests whether the halfedge, and its vertices, are used by things other than the
        faces, halfedges, and vertices passed in as the candidate set """
        if candidate_set is None:
            candidate_set = set()
        assert(isinstance(candidate_set, set))
        if self.twin is not None:
            candidates_plus_self = candidate_set.union([self, self.twin])
        else:
            candidates_plus_self = candidate_set.union([self])
        is_constrained = self.face is not None and self.face not in candidates_plus_self
        if self.origin is not None:
            is_constrained = is_constrained \
                            or self.origin.has_constraints(candidates_plus_self)
        if self.twin is not None:
            if self.twin.face is not None:
                is_constrained = is_constrained or self.twin.face not in candidates_plus_self
            if self.twin.origin is not None:
                is_constrained = is_constrained \
                                or self.twin.origin.has_constraints(candidates_plus_self)
        return is_constrained

    def is_infinite(self):
        """ If a halfedge has only one defined point, it stretches
            off into infinity """
        return self.origin is None or self.twin is None or self.twin.origin is None

    def connections_align(self, other):
        """ Verify that this and another halfedge's together form a full edge """
        assert(isinstance(other, HalfEdge))
        if self.twin.origin is None or other.origin is None:
            raise Exception("Invalid connection test")

        return self.twin.origin == other.origin

    def is_constrained(self):
        """ Check whether the edge has been forced within a bbox or circle"""
        return self.constrained or self.twin.constrained

    def set_constrained(self):
        """ Mark the full edge as forced within a bbox or circle """
        self.constrained = True
        self.twin.constrained = True

    def within(self, bbox):
        """ Check that both points in an edge are within the bbox """
        assert(isinstance(bbox, np.ndarray))
        assert(len(bbox) == 4)
        return self.origin.within(bbox) and self.twin.origin.within(bbox)

    def within_circle(self, centre, radius):
        """ is this halfedge within a given circle? """
        points = self.to_array()
        return in_circle(centre, radius, points)

    def outside(self, bbox):
        """ is this halfedge outside a given bbox? """
        verts = [x for x in self.get_vertices() if x is not None]
        return all([x.outside(bbox) for x in verts])

    def to_constrained(self, bbox):
        """ get the coords of the half-edge to within the
            bounding box of [min_x, min_y, max_x, max_y]
        """
        assert(self.origin is not None)
        assert(self.twin is not None)
        assert(self.twin.origin is not None)

        #Convert to an actual line representation, for intersection
        logging.debug("Constraining {} - {}".format(self.index, self.twin.index))
        as_line = Line.new_line(self.to_array())
        return as_line.constrain(*bbox)

    def swap_faces(self):
        """ Swap the registered face between the halfedges, to keep the halfedge
        as the external boundary of the face, and ordered ccw  """
        assert(self.face is not None)
        assert(self.twin is not None)
        #assert(self.twin.face is not None)
        origin_faces = self.face
        twin_faces = self.twin.face
        origin_faces.remove_edge(self)
        if twin_faces is not None:
            twin_faces.remove_edge(self.twin)
            twin_faces.add_edge(self)
        origin_faces.add_edge(self.twin)


    #------------------------------
    # def Vertex Access
    #------------------------------

    def add_vertex(self, vertex):
        """ Place a vertex into the first available slot of the full edge """
        assert(isinstance(vertex, Vertex))
        if self.origin is None:
            self.origin = vertex
            self.origin.register_half_edge(self)
        elif self.twin.origin is None:
            self.twin.origin = vertex
            self.twin.origin.register_half_edge(self.twin)
        else:
            raise Exception("trying to add a vertex to a full edge")

    def clear_vertices(self):
        """ remove vertices from the edge, clearing the vertex->edge references as well   """
        v1 = self.origin
        v2 = None
        self.origin = None
        if self.twin is not None:
            v2 = self.twin.origin
            self.twin.origin = None

        if v1 is not None:
            logging.debug("Clearing vertex {} from edge {}".format(v1.index, self.index))
            v1.unregister_half_edge(self)
        if v2 is not None:
            logging.debug("Clearing vertex {} from edge {}".format(v2.index, self.twin.index))
            v2.unregister_half_edge(self.twin)

    def replace_vertex(self, new_vert):
        """ Replace the vertex of this halfedge with a new one, unregistering the old """
        assert(isinstance(new_vert, Vertex))
        self.origin.unregister_half_edge(self)
        self.origin = new_vert
        self.origin.register_half_edge(self)

    def get_vertices(self):
        """ Get a tuple of the vertices of this halfedge """
        if self.twin is None:
            return (self.origin, None)
        return (self.origin, self.twin.origin)

    def to_array(self):
        """ Get an ndarray of the bounds of the edge """
        return np.row_stack((self.origin.to_array(), self.twin.origin.to_array()))

    def get_closer_and_further(self, centre):
        """ Return the edge vertices ordered to be [nearer, further] from a point,
        with a flag of whether the points have been switched from the edge ordering """
        assert(isinstance(centre, np.ndarray))
        distances = get_distance_raw(centre, self.to_array())
        if distances[0] < distances[1]:
            return (self.origin, self.twin.origin)
        else:
            return (self.twin.origin, self.origin)


    #------------------------------
    # def Edge Sequencing
    #------------------------------

    def add_next(self, next_edge, force=False):
        """ Add a halfedge to be the next in the chain """
        assert(next_edge is None or isinstance(next_edge, HalfEdge))
        if not force:
            assert(self.next is None)
            assert(next_edge is None or next_edge.prev is None)
        if self.next is not None:
            self.next.prev = None
        self.next = next_edge
        if self.next is not None:
            self.next.prev = self

    def add_prev(self, prev_edge, force=False):
        """ Set the half edge prior to this one in the CCW ordering """
        assert(prev_edge is None or isinstance(prev_edge, HalfEdge))
        if not force:
            assert(self.prev is None)
            assert(prev_edge is None or prev_edge.next is None)
        if self.prev is not None:
            self.prev.next = None
        self.prev = prev_edge
        if self.prev is not None:
            self.prev.next = self

    def connect_next_to_prev(self):
        """ Removes this Halfedge from the ordering """
        hprev = self.prev
        hnext = self.next
        if hprev is not None:
            hprev.next = hnext
        if hnext is not None:
            hnext.prev = hprev

    #------------------------------
    # def Cleanup
    #------------------------------

    def mark_for_cleanup(self):
        """ Marks this halfedge for cleanup. NOT for the twin,
        due to degenerate cases of hedges at boundaries """
        self.marked_for_cleanup = True



"""
How to convert ASTs to an external format (eg: Clingo)
"""
##-- imports
from __future__ import annotations

from importlib.resources import files
import abc
import logging as logmod
import os
from string import Template
from dataclasses import InitVar, dataclass, field
from pathlib import Path

##-- end imports

logging = logmod.getLogger(__name__)

from jgdv.apis.clingo.ast import SolverAST

@dataclass
class Compiler_i(metaclass=abc.ABCMeta):
    """
    Interface for compiling InstaASTR down to a
    specific solver format
    """
    _compiled_text : list[str] = field(init=False, default_factory=list)
    # TODO add delayed registers

    @property
    def compilation(self):
        return "\n".join(self._compiled_text)

    def clear(self):
        self._compiled_text = []

    def insert(self, pattern:str|Template, **kwargs):
        """
        insert a given pattern text into the compiled output,
        formatting it with kwargs.
        """
        match pattern:
            case Template():
                self._compiled_text.append(pattern.safe_substitute(kwargs))
            case str() if not bool(kwargs):
                self._compiled_text.append(pattern)
            case str():
                self._compiled_text.append(pattern.format_map(kwargs))
            case _:
                raise TypeError("Unrecognised compile pattern type", pattern)

    def load_prelude(self) -> str:
        return ""

    @abc.abstractmethod
    def compile(self, data:list[SolverAST]) -> str: pass



"""
Provides broad drawing support for a dcel
"""
#pylint: disable=too-many-arguments
import logging as root_logger
from ..constants import WIDTH
from ..drawing import clear_canvas
from .constants import FACE_COLOUR, EDGE_COLOUR, VERT_COLOUR, BACKGROUND_COLOUR


logging = root_logger.getLogger(__name__)

class DCELDraw:

    @staticmethod
    def draw(ctx, dcel, settings:DCELDrawSettings):
        """ A top level function to draw a dcel  """

        clear_canvas(ctx, colour=settings.background, bbox=dcel.bbox)

        if settings.faces:
            ctx.set_source_rgba(*face_colour)
            DCELDraw._faces(ctx, dcel, text=text)

        if settings.edges:
            ctx.set_source_rgba(*edge_colour)
            DCELDraw._edges(ctx, dcel, text=text, width=edge_width)

        if settings.verts:
            ctx.set_source_rgba(*vert_colour)
            DCELDraw._vertices(ctx, dcel)

    @staticmethod
    def _faces(ctx, dcel, text=True, clear=False):
        """ Draw ll faces in a dcel """
        for f in dcel.faces:
            f.draw(ctx, clear=clear, text=text)

    @staticmethod
    def _edges(ctx, dcel, text=True, width=WIDTH):
        """ Draw all edges in a dcel """
        original_text_state = text
        drawn_texts = set()
        for edge in dcel.halfEdges:
            if edge.index not in drawn_texts:
                text = original_text_state
                drawn_texts.add(edge.index)
                drawn_texts.add(edge.twin.index)
            else:
                text = False
            edge.draw(ctx, text=text, width=width)

    @staticmethod
    def _vertices(ctx, dcel):
        """ Draw all the vertices in a dcel as dots """
        for v in dcel.vertices:
            v.draw(ctx)

def draw_dcel(ctx, dcel, text=False, faces=True, edges=False, verts=False,
              face_colour=None, edge_colour=None, vert_colour=None,
              background_colour=None, edge_width=WIDTH):
    """ A top level function to draw a dcel  """
    if face_colour is None:
        face_colour = FACE_COLOUR
    if edge_colour is None:
        edge_colour = EDGE_COLOUR
    if vert_colour is None:
        vert_colour = VERT_COLOUR
    if background_colour is None:
        background_colour = BACKGROUND_COLOUR

    clear_canvas(ctx, colour=background_colour, bbox=dcel.bbox)

    if faces:
        ctx.set_source_rgba(*face_colour)
        draw_dcel_faces(ctx, dcel, text=text)

    if edges:
        ctx.set_source_rgba(*edge_colour)
        draw_dcel_edges(ctx, dcel, text=text, width=edge_width)

    if verts:
        ctx.set_source_rgba(*vert_colour)
        draw_dcel_vertices(ctx, dcel)

def draw_dcel_faces(ctx, dcel, text=True, clear=False):
    """ Draw ll faces in a dcel """
    for f in dcel.faces:
        f.draw(ctx, clear=clear, text=text)

def draw_dcel_edges(ctx, dcel, text=True, width=WIDTH):
    """ Draw all edges in a dcel """
    original_text_state = text
    drawn_texts = set()
    for edge in dcel.halfEdges:
        if edge.index not in drawn_texts:
            text = original_text_state
            drawn_texts.add(edge.index)
            drawn_texts.add(edge.twin.index)
        else:
            text = False
        edge.draw(ctx, text=text, width=width)

def draw_dcel_vertices(ctx, dcel):
    """ Draw all the vertices in a dcel as dots """
    for v in dcel.vertices:
        v.draw(ctx)



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@dataclass
class LineSweepResult:
    """ Utility class to collect the start and end sets of halfedges,
    along with the set of intersections contained within halfedges,
    and the vertex describing that point.

    Starting and Ending are defined from Top -> Bottom, Left -> Right
    """

    vertex  : Vertex   = field()
    start   : HalfEdge = field()
    contain : Any      = field()
    end     : HalfEdge = field()

    def __repr__(self):
        return f"<Intersection: {self.vertex}\nStart: {self.start}\nContain: {self.contain}\nEnd: {self.end}>"



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

@dataclass
class Leaf:

    data : List[Any] = field(default_factory=list)

    __repr__ = __str__

    def __len__(self):
        return len(self.data)

    def __str__(self):
        return "Leaf Group({})".format(len(self))

    def get_tuple_list(self):
        return [x.to_tuple() for x in self.data]

    def insert(self, name, url, tags, query_dict, full_path):
        new_leaf = LeafComponent(name, url, tags, full_path, query_dict)
        if new_leaf in self.data:
            logging.info("Merging tags")
            existing = self.data.index(new_leaf)
            existing.tags.update(tags)
            return existing
        else:
            self.data.append(new_leaf)
            return new_leaf

    def filter_queries(self, query_set):
        for x in self.data:
            x.filter_queries(query_set)

@dataclass
class LeafComponent:

    name      : str            = field()
    url       : str            = field()
    tags      : List[str]      = field()
    full_path : str            = field()
    query     : Dict[Any, Any] = field(default_factory=dict)

    __repr__ = __str__

    def __eq__(self, other):
        if not isinstance(other, LeafComponent):
            return False
        if self.full_path != other.full_path:
            return False
        return True

    def __str__(self):
        return "Leaf({})".format(self.full_path)

    def filter_queries(self, query_set):
        for k in list(self.query.keys()):
            if k in query_set:
                del self.query[k]

    def reconstruct(self, key=None):
        copied = {}
        copied.update(self.query)
        if key in copied:
            del copied[key]
        query_str = urlencode(copied, True)
        full_path = urlunparse((self.url.scheme,
                                self.url.netloc,
                                self.url.path,
                                self.url.params,
                                query_str,
                                self.url.fragment))
        return full_path

    def to_tuple(self):
        return Bookmark(self.reconstruct(),
                        self.tags,
                        name=self.name)



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import numpy as np

def quantize(matrix:np.ndarray, layers:int):
    """ Given a basic heightmap, quantize it """
    logging.info("Quantizing")
    #pad the size a bit
    qmatrix       = np.zeros(matrix.shape)
    scaling       = 100
    scaling_recip = 1 / scaling
    quantize_base = int(scaling / layers)
    scaled_matrix = matrix * scaling

    for l in range(1,layers):
        quantize_value =  quantize_base * l
        qmatrix        += scaled_matrix > quantize_value

    qmatrix *= quantize_base
    qmatrix *= scaling_recip
    return qmatrix



# pylint: disable=bad-whitespace,invalid-name,line-too-long
#
##-- imports
from __future__ import annotations

import logging as logmod

import pyparsing as pp
##-- end imports

logging = logmod.getLogger(__name__)

COMMENT_RE       = ""
WORD_COMPONENT_S = ""
WHITE_SPACE      = ""
TAB_S            = ""

pp.ParserElement.set_default_whitespace_chars(WHITE_SPACE)

s         = pp.Suppress
op        = pp.Optional
orm       = pp.OneOrMore
zrm       = pp.ZeroOrMore

s_lit = lambda x: s(pp.Literal(x))
s_key = lambda x: s(pp.Keyword(x))

COMMENT           = pp.Regex(COMMENT_RE)
END               = s_key(DSYM.END)
END.set_whitespace_chars(" \n")

ln                = s(pp.White("\n\r", max=1).set_whitespace_chars("\t "))
manyLine          = s(pp.White("\n\r", min=1).set_whitespace_chars("\t "))
emptyLine         = s(ln + manyLine)
opLn              = op(ln)
tab               = pp.White(TAB_S, min=2).set_whitespace_chars("\r\n")

emptyLine.set_name("emptyLine")
ln.set_name("line")
opLn.set_name("OptionalLine")
tab.set_name("tab")

def NG(name, grp):
    """ Name and Group """
    return pp.Group(grp).set_results_name(name)

def N(name, parser):
    return parser.set_results_name(name)

def gap_fail_action(s, loc, expr, err):
    logging.warning("{}\n{}".format(str(err), err.markInputline()))


# Basic Syntax
ARROW     = s_lit('->')
AT        = s_lit('@')
BSLASH    = s_lit('\\')
CBRACKET  = s_lit(']')
COLON     = s_lit(':')
COMMA     = s_lit(',')
CPAR      = s_lit(')')
DASH      = s_lit('-')
DBLARROW  = s_lit('=>')
DBLCOLON  = s_lit("::")
DOLLAR    = s_lit('$')
DOUBLEBAR = s_lit('||')
HASH      = s_lit("#")
LESS      = s_lit('<')
MORE      = s_lit('>')
OBRACKET  = s_lit('[')
OPAR      = s_lit('(')
QMARK     = s_lit('?')
SLASH     = s_lit('/')
TILDE     = s_lit('~')
VBAR      = s_lit('|')
# DELIM     = ~END + (COMMA | (ln + tab))
DELIM    = (ln ^ COMMA).set_whitespace_chars(" \t")

COLLAPSE_CONTEXT = s_lit(DSYM.COLLAPSE_CONTEXT)

FUNC_SYMBOL      = s(pp.Word(DSYM.FUNC))

BIND             = s_lit(DSYM.BIND)
AT_BIND          = s_lit(DSYM.AT_BIND)

QUERY            = s_lit(DSYM.QUERY)
TAG              = s_lit(DSYM.TAG)

NEGATION         = pp.Literal(DSYM.NEGATION)(DS.NEGATION)

gap              = emptyLine #.set_fail_action(gap_fail_action)
component_gap    = s(orm(emptyLine))
file_cruft       = pp.ZeroOrMore(ln)

END.set_name("End")
file_cruft.set_name("EmptyLines")
gap.set_name("EmptyLine")
COLON.set_name("Colon")
DELIM.set_name("Delimiter")



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
# import more_itertools as mitz
# from boltons import
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class JGDVMultiKey(JGDVBaseKey):
    """ A string or path of multiple keys """

    def __init__(self, val:str|pl.Path):
        self.value : str|pl.Path        = val
        self._keys : set[DootSimpleKey] = set(DootSimpleKey(x) for x in PATTERN.findall(str(val)))

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return "<DootMultiKey: {}>".format(str(self))

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        match other:
            case JGDVBaseKey() | str() | pl.Path():
                return str(self) == str(other)
            case _:
                return False

    def keys(self) -> set(DootSimpleKey):
        return self._keys

    @property
    def form(self):
        """ Return the key in its use form """
        return str(self)

    def expand(self, spec=None, state=None, *, rec=False, insist=False, chain:list[JGDVBaseKey]=None, on_fail=Any, locs=None, **kwargs) -> str:
        try:
            return DootFormatter.fmt(self.value, _spec=spec, _state=state, _rec=rec, _insist=insist, _locs=locs)
        except (KeyError, TypeError) as err:
            if bool(chain):
                return chain[0].expand(spec, state, rec=rec, chain=chain[1:], on_fail=on_fail)
            elif on_fail != Any:
                return on_fail
            else:
                raise err

    def within(self, other:str|dict|TomlGuard) -> bool:
        return str(self) in other

    def to_type(self, spec, state, type_=Any, chain:list[JGDVBaseKey]=None, on_fail=Any) -> Any:
        raise TypeError("Converting a MultiKey to a type doesn't build sense", self)



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import sys
import trace

class TraceHelper:
    """ Utility to simplify using the trace library, as a context manager

      see https://docs.python.org/3/library/trace.html
    """
    def __init__(self):
        self._tracer = trace.Trace(ignoredirs=[sys.exec_prefix], count=1, countfuncs=0, countcallers=0)
        self._results = None
        self._write_to = None

    def __enter__(self) -> Any:
        #return self
        return

    def __exit__(self, exc_type, exc_value, exc_traceback) -> bool:
        self._results = self._tracer.results()
        if self._write_to:
            self._results.write_results(summary=True, show_missing=True, coverdir=self._write_to)
       # return False to reraise errors
        return

    def this(self, func):
        """ Run a Trace on the passed in function """
        self.runfunc(func)



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import gzip
import pickle
from time import time

import scrapy
from itemadapter import ItemAdapter, is_item
from scrapy import signals
from scrapy.crawler import CrawlerProcess
from scrapy.dupefilters import RFPDupeFilter
from scrapy.exporters import XmlItemExporter
from scrapy.http import Headers, Response
from scrapy.http.request import Request
from scrapy.responsetypes import responsetypes
from scrapy.spiders import Spider
from scrapy.utils.httpobj import urlparse_cached
from w3lib.http import headers_dict_to_raw, headers_raw_to_dict


class CachePolicy:
    """
    Defines when cache's become invalid
    """

    def __init__(self, settings):
        self.ignore_schemes = settings.getlist("HTTPCACHE_IGNORE_SCHEMES")
        self.ignore_http_codes = [
            int(x) for x in settings.getlist("HTTPCACHE_IGNORE_HTTP_CODES")
        ]

    def should_cache_request(self, request):
        return urlparse_cached(request).scheme not in self.ignore_schemes

    def should_cache_response(self, response, request):
        return response.status not in self.ignore_http_codes

    def is_cached_response_fresh(self, cachedresponse, request):
        return True

    def is_cached_response_valid(self, cachedresponse, response, request):
        return True

class CacheStorage:
    """
    Defines how responses are stored and retrieved in the cache
    """

    def __init__(self, settings):
        self.cachedir        = settings["HTTPCACHE_DIR"]
        self.expiration_secs = settings.getint("HTTPCACHE_EXPIRATION_SECS")
        self.use_gzip        = settings.getbool("HTTPCACHE_GZIP")
        self._open           = gzip.open if self.use_gzip else open

    def open_spider(self, spider: Spider):
        logging.debug("Using filesystem cache storage in %s", self.cachedir, extra={"spider": spider},)
        self._fingerprinter = spider.crawler.request_fingerprinter

    def close_spider(self, spider):
        pass

    def retrieve_response(self, spider: Spider, request: Request):
        """Return response if present in cache, or None otherwise."""
        metadata       = self._read_meta(spider, request)
        if metadata is None:
            return  # not cached
        logging.debug("Reusing cached response")
        rpath          = self._get_request_path(spider, request)
        with self._open(rpath / "response_body", "rb") as f:
            body       = f.read()
        with self._open(rpath / "response_headers", "rb") as f:
            rawheaders = f.read()
        url            = metadata.get("response_url")
        status         = metadata["status"]
        headers        = Headers(headers_raw_to_dict(rawheaders))
        respcls        = responsetypes.from_args(headers=headers, url=url, body=body)
        response       = respcls(url=url, headers=headers, status=status, body=body)
        return response

    def store_response(self, spider: Spider, request: Request, response):
        """Store the given response in the cache."""
        rpath = self._get_request_path(spider, request)
        if not rpath.exists():
            rpath.mkdir(parents=True)
        metadata = {
            "url": request.url,
            "method": request.method,
            "status": response.status,
            "response_url": response.url,
            "timestamp": time(),
        }

        (rpath / "meta").write_bytes(repr(metadata).encode())
        (rpath / "pickled_meta").write_bytes(pickle.dumps(metadata, protocol=4))
        (rpath / "response_headers").write_bytes(headers_dict_to_raw(response.headers))
        (rpath / "response_body").write_bytes(response.body)
        (rpath / "request_headers").write_bytes(headers_dict_to_raw(request.headers))
        (rpath / "request_body").write_bytes(request.body)

    def _get_request_path(self, spider: Spider, request: Request) -> pl.Path:
        key = self._fingerprinter.fingerprint(request).hex()
        return pl.Path(self.cachedir, spider.name, key[0:2], key)

    def _read_meta(self, spider: Spider, request: Request):
        rpath = self._get_request_path(spider, request)
        metapath = rpath / "pickled_meta"
        if not metapath.exists():
            return  # not found
        mtime = metapath.stat().st_mtime
        if 0 < self.expiration_secs < time() - mtime:
            return  # expired
        with self._open(metapath, "rb") as f:
            return pickle.load(f)


class CacheDupeFilter(RFPDupeFilter):
    """
    Filter Requests by cache status

    """

    def request_seen(self, request):
        logging.debug("Request Seen: %s", request.url)
        # fp = self.request_fingerprint(request)
        return super().request_seen(request)



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class RBTreeUtils:

    def min(self):
        """ Get the min value of the tree """
        if self.root is None:
            return None
        return self.root.min()

    def max(self):
        """ Get the max value of the tree """
        if self.root is None:
            return None
        return self.root.max()

    def search(self, value, cmp_func=None, eq_func=None, cmp_data=None, closest=False, start=None):
        """ Search the tree for a value """
        logging.debug("Searching for: {}".format(value))

        if cmp_func is None:
            cmp_func = self.cmp_func
        if eq_func is None:
            eq_func = self.eq_func
        if start is None:
            start = self.root
        parent = start
        current = start
        comp = None
        while current is not None and not eq_func(current, value, cmp_data):
            parent = current
            comp = cmp_func(current, value, cmp_data)
            if comp is Directions.LEFT:
                logging.debug("left")
                current = current.left
            else:
                assert(comp is Directions.RIGHT)
                logging.debug("right")
                current = current.right

        if closest and current is None:
            #closest non-exact match found
            logging.debug("Found Closest: {}, {}".format(parent, comp))
            return (parent, comp)
        elif current is None:
            #nothing found
            logging.debug("Found Nothing")
            return (None, None)
        else:
            #exact match found
            logging.debug("Found Exact: {}, {}".format(current, comp))
            return (current, comp)

    def update_values(self, func, func_data):
        """ Call a function on all stored values,
        function signature:  f(value, func_data)
        """
        assert(isinstance(func, (FunctionType, partial)))
        for node in self.nodes:
            func(node.value, func_data)

    def insert(self, *args, data=None, cmp_data=None):
        """ Insert data into the tree """
        nodes = []
        for x in args:
            new_node = self.__insert(x, data=data, cmp_data=cmp_data)
            nodes.append(new_node)
        return nodes

    def delete(self, *args, cleanup_func=None):
        """ Delete a value from the tree """
        if cleanup_func is None:
            cleanup_func = self.cleanup_func
        to_remove = set(args)
        while bool(to_remove):
            target = to_remove.pop()
            logging.debug("Removing Target: {}".format(target))
            assert(isinstance(target, Node))
            if target not in self.nodes:
                continue
            to_remove.update(cleanup_func(target))
            rb_tree_delete(self, target)
            self.nodes.remove(target)

    def delete_value(self, *args, cmp_func=None, eq_func=None, cleanup_func=None, cmp_data=None):
        """ Delete a value from the tree, and rebalance  """
        for val in args:
            node, _ = self.search(val, cmp_func=cmp_func,
                                  eq_func=eq_func, cmp_data=cmp_data)
            if node is not None:
                self.delete(node, cleanup_func=cleanup_func)
            else:
                logging.debug("Not Found: {}".format(val))

    def __insert(self, value, data=None, cmp_data=None):
        """ Insert a value into the tree """
        parent, direction = self.search(value, closest=True, cmp_data=cmp_data)
        if direction is Directions.LEFT:
            return self.insert_predecessor(parent, value, data=data)
        else:
            return self.insert_successor(parent, value, data=data)

    def __balance(self, node):
        assert(isinstance(node, Node))
        rb_tree_fixup(self, node)

    def count_black_height(self, node=None):
        """ Given a node, count all paths and check they have the same black height """
        if node is None:
            if self.root is None:
                return None
            node = self.root
        stack = [node]
        leaves = []
        while bool(stack):
            current = stack.pop()
            if current.left is None and current.right is None:
                leaves.append(current)
            else:
                if current.left is not None:
                    stack.append(current.left)
                if current.right is not None:
                    stack.append(current.right)

        all_heights = [x.getBlackHeight(node) for x in leaves]
        return all_heights



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import networkx as nx
from jgdv.files.org.file import OrgStrBuilder
from jgdv.files.twitter.writers.util import parse_date

media_dict : Final[Callable] = lambda: defaultdict(list)

@dataclass
class OrgThreadWriter:
    """ Given a thread object, build a string representation for it
    redirecting absolute media paths to relative
    """

    user          : str
    date          : datetime.datetime
    tags          : set[str]
    main          : list[TwitterTweet]       = field(default_factory=list)
    conversations : list[List[TwitterTweet]] = field(default_factory=list)
    uses          : set[str]                 = field(default_factory=set)
    links         : set[str]                 = field(default_factory=set)
    media         : dict                     = field(default_factory=media_dict)

    redirect_url  : str = "{}_files"
    date_re       : str = r"%a %b %d %H:%M:%S +0000 %Y"
    tag_pattern   : str = ":{}:"
    tag_sep       : str = ":"
    tag_col       : int = 80
    file_link_pattern : str = "[[file:./{}][{}]]"

    @staticmethod
    def build(thread:dict, tweets:list, users:list, source_tags:dict[str, set[str]]) -> OrgThreadWriter:
        logging.info("Creating thread")

        main_thread_ids = thread['main_thread']
        sub_thread_ids  = thread['rest']
        quote_ids       = thread['quotes']
        base_user       = thread["base_user"]
        tweet_lookup    = {x['id_str'] : x for x in tweets}
        tags            = {y.strip() for x in main_thread_ids for y in source_tags[x] if x in source_tags}
        users_lookup    = {x['id_str'] : x for x in users}

        min_thread_date = min(((x,parse_date(tweet_lookup[x]['created_at'], self.date_re))
                              for x in main_thread_ids if x in tweet_lookup), key=lambda v:v[1])
        thread_date = min_thread_date[1]

        obj = OrgThreadWriter(base_user, thread_date, tags)

        # add tweets of main thread
        for x in main_thread_ids:
            if x not in tweet_lookup:
                obj.main.append(TwitterTweet("null", base_user))
                continue

            obj.add_use(x)
            tweet_obj = TwitterTweet.build(base_user, tweet_lookup[x], users_lookup, tweet_lookup)
            obj.main.append(tweet_obj)
            for key, values in tweet_obj.media.items():
                obj.media[key] += values

            obj.links.update(tweet_obj.links)

        # Add sub conversations
        for conv in sub_thread_ids:
            conv_list = []
            for x in conv:
                if x not in tweet_lookup:
                    conv_list.append(TwitterTweet("null", base_user, level=5))
                    continue
                obj.add_use(x)
                tweet_obj = TwitterTweet.build(base_user, tweet_lookup[x], users_lookup, tweet_lookup, level=5)
                conv_list.append(tweet_obj)
                obj.media.update(tweet_obj.media)
                obj.links.update(tweet_obj.links)

            obj.conversations.append(conv_list)

        return obj

    def add_use(self, value):
        """
        Record a use of a tweet id
        """
        self.uses.add(value)

    @staticmethod
    def retarget_url(base, url):
        base_p = pl.Path(OrgThreadWriter.redirect_url.format(base))
        url_p  = pl.Path(url)
        logging.debug("Retargeting URL: %s", url_p.name)
        return str(base_p / url_p.name)


    def __str__(self):
        output = OrgStrBuilder()

        heading_str = f"Thread: {self.date}"
        tags_str = ""
        if bool(self.tags):
            tags_str = self.tag_pattern.format(self.tag_sep.join(self.tags))

        tag_pad = max(0, self.tag_col - len(heading_str))
        output.heading(2, f"{heading_str}{tag_pad*' '}{tags_str}")
        output.heading(3, "Main Thread")
        output.add(*self.main)

        output.heading(3, "Conversations: ", str(len(self.conversations)))
        for conv in self.conversations:
            if not bool(conv):
                continue

            output.heading(4, conv[0].at)
            output.add(*conv)

        output.heading(3, "Links: ", str(len(self.links)))
        output.links(self.links)
        output.nl

        output.heading(3, "Media: ", str(len(self.media)))
        retarget = lambda x: (OrgThreadWriter.retarget_url(self.user, x['url']), pl.Path(x['url']).name)
        media_urls = []
        media_urls += [retarget(media) for media in self.media.get('photo', [])]
        media_urls += [retarget(media) for media in self.media.get('video', [])]

        media_local =[self.file_link_pattern.format(*x) for x in media_urls]
        output.add(*media_local)
        output.nl

        return str(output)



#!/usr/bin/env python3
##-- imports
from __future__ import annotations
import logging as logmod

##-- end imports

logging = logmod.getLogger(__name__)

config = acab.config

Node          = DI.StructView
Value         = AT.Value
Structure     = AT.DataStructure
Engine        = AT.Engine
Contexts      = AT.CtxSet

class FlattenBreadthTrieSemantics(basic.StructureSemantics, SI.StructureSemantics_i):
    """
    Trie Semantics which map values -> Nodes
    Searches *Breadth First*

    Has sentence-flattening logic for querying.
    """

    def verify(self, instruction) -> bool:
        return isinstance(instruction, Sentence)

    def compatible(self, struct):
        is_bns = isinstance(struct, BasicNodeStruct)
        has_all_node_comp = "all_nodes" in struct.components
        return is_bns or has_all_node_comp

    def insert(self, sen, struct, *, data=None, ctxs=None):
        # TODO can insert handle @vars?
        data = data or {}

        data.update({'source_semantics': self})
        logging.info("Inserting: {!r} into {!r}", sen, struct)
        # Get the root
        current = self.lookup()[0].up(struct.root)
        for word in sen:
            spec = self.lookup(current)
            accessible = spec[0].access(current, word, data=data)
            if bool(accessible):
                current = accessible[0]
            else:
                next_spec = self.lookup(word)
                new_node = next_spec[0].make(word, data=data)
                struct.components['all_nodes'][new_node.uuid] = new_node
                current = spec[0].insert(current, new_node, data=data)

        return current

    def delete(self, sen, struct, *, data=None, ctxs=None):
        logging.debug("Removing: {!r} from {!r}", sen, struct)

        pos_sen = sen.copy(data={DS.NEGATION:False})
        results = self.query(pos_sen, struct, data=data, ctxs=ctxs)

        for ctx in results:
            assert(ctx.current_node is not None)
            assert(ctx.current_node.node.parent is not None)
            # At leaf:
            # remove current from parent
            current  = ctx.current_node.node
            parent   = current.parent()
            spec     = self.lookup(parent)
            spec[0].remove(parent, current.value, data=data)

    def query(self, sen, struct, *, data=None, ctxs=None):
        """ Breadth First Search Query """
        assert(ctxs is not None)
        logging.info("Querying: {!r} : {!r}", sen ,struct)
        clause_flatten = DS.FLATTEN not in sen.data or sen.data[DS.FLATTEN]
        root           = DI.StructView(struct.root, self) if isinstance(struct, DI.Structure_i) else struct
        assert(isinstance(root, Node))

        cqm = FlattenQueryManager(sen, root, ctxs)
        with cqm:
            for source_word in cqm.current:
                for bound_word, ctxInst, view in cqm.active:
                    # This happens here to catch when $x -> a.sub.sentence
                    # in a.larger.query.$x.blah?
                    should_flatten = clause_flatten and bound_word and (DS.FLATTEN not in bound_word.data or bound_word.data[DS.FLATTEN])
                    if isinstance(bound_word, Sentence_i) and should_flatten:
                        # sub query when dealing with a sentence that needs to be flattened
                        subctx = self.query(bound_word, view, data=data, ctxs=ctxs.subctx([ctxInst.uuid]))
                        for sub_inst in subctx.active_list(clear=True):
                            subctx.push(sub_inst.progress(source_word, [sub_inst.current_node]))
                        cqm.queue_ctxs(subctx.active_list())
                    elif source_word.is_at_var and not bool(cqm._current_constraint):
                        continue
                    elif source_word.is_at_var:
                        assert(source_word == sen[0])
                        cqm.maybe_test([view])
                    else:
                        if not isinstance(view, DI.StructView):
                            breakpoint()
                        spec = self.lookup(view.node)
                        results = spec[0].access(view.node,
                                                 bound_word,
                                                 data=data)

                        cqm.maybe_test([DI.StructView(x, self) for x in results])

        return cqm.finished

    def to_sentences(self, struct, *, data=None, ctxs=None):
        """ Convert a trie to a list of sentences
        essentially a dfs of the structure,
        ensuring only leaves are complex structures.

        structures are converted to words for use within sentences
        """
        # TODO if passed a node, use that in place of root
        result_list = []
        # Queue: list[Tuple[list[Value], Node]]

        match struct:
            case DI.Structure_i():
                root = struct.root
            case DI.Node_i():
                root = struct.Root()
                root.add(struct)
            case DI.StructView():
                root = struct.node.Root()
                root.add(struct.node)
            case _:
                raise TypeError("Unknown struct passed in")

        # TODO add depth limit
        queue = [([], root)]
        while bool(queue):
            path, current = queue.pop(0)
            updated_path  = path + [current.value]
            spec          = self.lookup(current)
            accessible    = spec[0].access(current, None, data=data)
            if bool(accessible):
                # branch
                queue += [(updated_path, x) for x in accessible]

            if not bool(accessible) or isinstance(current.value, Instruction):
                # Leaves and Statements
                # Always ignore the root node, so starting index is 1
                words = [x.to_word() if not isinstance(x, Sentence_i) else x for x in updated_path[1:-1]]
                words.append(updated_path[-1])
                result_list.append(Sentence(words))

        return result_list

    @staticmethod
    def from_sentences(self, sens):
        raise NotImplementedError()

    def to_word(self):
        raise NotImplementedError()



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from typing import Self
from abc import abstractmethod
from tomlguard import TomlGuard
from importlib.metadata import EntryPoint

@runtime_checkable
class PluginLoader_p(Protocol):
    """ Base for the first things loaded: plugins."""
    loaded : ClassVar[TomlGuard] = None

    @staticmethod
    def get_loaded(group:str, name:str) -> None|str:
        if PluginLoader_p.loaded is None:
            return None
        if group not in PluginLoader_p.loaded:
            return None
        matches = [x.value for x in PluginLoader_p.loaded[group] if x.name == name]
        if bool(matches):
            return matches[0]

        return None

    @abstractmethod
    def setup(self, extra_config:TomlGuard) -> Self:
        raise NotImplementedError()

    @abstractmethod
    def load(self) -> TomlGuard[EntryPoint]:
        raise NotImplementedError()




#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import asyncio
import struct

# BigEnd, continue?, msglen
head = struct.Struct(">?L")

async def tcp_echo_client(message):
    reader, writer = await asyncio.open_connection('127.0.0.1', 8888)

    print(f'Sent: {message!r}')
    b_message = message.encode()
    writer.write(head.pack(True, len(b_message)))
    writer.write(b_message)
    await writer.drain()

    while True:
        data = await reader.read(head.size)
        match head.unpack(data):
            case True, size:
                print("Reading: ", size)
                data  = await reader.read(size)
                print("Message: ", data.decode())
                writer.write(head.pack(True, len(b_message)))
                writer.write(b_message)
                await writer.drain()
                await asyncio.sleep(1)
            case False, size:
                response = head.pack(False, 0)
                writer.write(response)
                await writer.drain()
                break
            case _:
                raise Exception("Bad Message Received")

    print('Close the connection')
    writer.close()
    await writer.wait_closed()

asyncio.run(tcp_echo_client('Hello World!'))



#!/usr/bin/env python3
# pylint: disable=no-member
##-- imports
from __future__ import annotations

import abc
import logging as logmod
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from os.path import splitext

##-- end imports

import pyparsing as pp

logging = logmod.getLogger(__name__)
SC      = SimpleLogColour

config = acab.config
import_dict = config.imports.specific

@register_class("report")
class ReportCmd:
    """
    Print reports about the system configurationz

    Prints all information by default,
    If you type `report {section}` prints only that section.

    Available sections:
    dsl
    engine
    module   / mod
    operator / ops
    printers
    semantics

    """

    def __init__(self):
        self._parser = self._gen_parser()

    def _gen_parser(self):
        operator_kw  = pp.MatchFirst([pp.Keyword("ops"),
                                      pp.Keyword("operator")])("operator")
        module_kw    = pp.MatchFirst([pp.Keyword("mod"),
                                      pp.Keyword("module")])("module")
        semantic_kw  = pp.Keyword("semantics")("semantics")
        printer_kw   = pp.Keyword("printers")("printers")
        dsl_kw       = pp.Keyword("dsl")("dsl")
        engine_kw    = pp.Keyword("engine")("engine")

        eol          = pp.line_end("all")

        report_parser = pp.ZeroOrMore(pp.MatchFirst([operator_kw,
                                                    module_kw,
                                                    semantic_kw,
                                                    printer_kw,
                                                    dsl_kw,
                                                    engine_kw,
                                                    eol]))
        return report_parser

    def __call__(self, line):
        self.do_report(line)

    def do_report(self, line):
        logging.info(f"Getting reports: {line}")
        try:
            params                              = self._parser.parse_string(line, parse_all=True)
        except pp.ParseException as err:
            print(f"Unrecognised Command detail: {err.pstr}")
            print(self.__doc__)
            return

        allow_all    : bool                 = (not bool(params)) or "all" in params
        self.modules : list[ModuleFragment] = self._repl.state.engine._module_loader.loaded
        # modules
        if allow_all or "module" in params:
            self.print_modules(params)

        if allow_all or "dsl" in params:
            self.print_dsl(params)

        if allow_all or "semantics" in params:
            self.print_semantics(params)

        if allow_all or "operator" in params:
            self.print_operators(params)

        if allow_all or "printers" in params:
            self.print_printers(params)

        if allow_all or "engine" in params:
            self.print_engine(params)

    def print_modules(self, params):
        ColPr.print_header("Modules")
        mod_target = self.modules
        if 'mod_target' in params:
            mod_target = [x for x in self.modules if x.source in params['mod_target']]

        for mod in mod_target:
            ColPr.print_module_colour(mod)
        print("--")
        print(f"Loaded Modules: {SC.green(len(self.modules))}")

    def print_dsl(self, params):
        ColPr.print_header("DSL")
        dsl = self._repl.state.engine._dsl
        ColPr.print_class_colour("DSL Base", dsl)
        ColPr.print_handler_system(dsl)

    def print_semantics(self, params):
        ColPr.print_header("Semantics")
        semantic = self._repl.state.engine.semantics
        ColPr.print_class_colour("Semantic Base", semantic)
        ColPr.print_handler_system(semantic)

        print("----------")
        mods_with_semantics = [x for x in self.modules if len(x.semantics) > 0]
        if bool(mods_with_semantics):
            print("Module Semantics: ")
            count = defaultdict(lambda: 0)
            for mod in mods_with_semantics:
                print(f"Module: {mod.source}")
                for frag in mod.semantics:
                    print(f"\t{frag}")

    def print_operators(self, params):
        ColPr.print_header("Operators")
        count = 0
        for mod in self.modules:
            count += len(mod.operators)
            for op in mod.operators:
                print("\t", ColPr.two_part_split(self._repl.state.engine.pprint(target=[op])))

        print("--")
        print(f"Loaded Operators: {SC.green(count)}")
        # TODO print Context Operator Bindings

    def print_printers(self, params):
        ColPr.print_header("Printers")
        printer = self._repl.state.engine.printer
        ColPr.print_class_colour("Printing Base", printer)
        ColPr.print_handler_system(printer)

        print("----------")
        mods_with_printers = [x for x in self.modules if len(x.printers) > 0]
        if bool(mods_with_printers):
            print("Module Printers: ")
            count = defaultdict(lambda: 0)
            for mod in mods_with_printers:
                print(f"Module: {mod.source}")
                for frag in mod.printers:
                    print(f"{frag}")

    def print_engine(self, params):
        ColPr.print_header("Engine")
        engine = self._repl.state.engine
        ColPr.print_class_colour("Active Engine", engine)
        load_paths = '\n\t'.join(engine.load_paths) or SC.red("None")
        print(f"\nLoad Paths:\n\t{load_paths}")

        print(f"\nSystem Wide Imports:")
        for k,v in sorted(import_dict().items()):
            key_str = SC.blue(f"{k:<20}")
            path_split = splitext(v)
            print(f"\t{key_str} : {path_split[0]}{SC.blue(path_split[1])}")



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class DCELProcessor:

    def calculate_quad_tree(self, subverts=None):
        """ Recalculate the quad tree with all vertices, or a subselection of vertices """
        self.vertex_quad_tree = pyqtree.Index(bbox=self.bbox)
        verts = self.vertices
        if subverts is not None:
            assert(all([isinstance(x, Vertex) for x in subverts]))
            verts = subverts
        for vertex in verts:
            self.vertex_quad_tree.insert(item=vertex, bbox=vertex.bbox())

    def push_quad_tree(self):
        """ Add another vertex quad tree on top of the stack """
        self.quad_tree_stack.append(self.vertex_quad_tree)
        self.vertex_quad_tree = pyqtree.Index(bbox=self.bbox)

    def pop_quad_tree(self):
        """ Remove the top vertex quad tree from the stack """
        assert(bool(self.quad_tree_stack))
        sub_layer = self.quad_tree_stack.pop()
        if self.should_merge_stacks:
            for x in self.vertex_quad_tree.intersect(self.bbox):
                if x not in sub_layer.intersect(x.bbox()):
                    sub_layer.insert(item=x, bbox=x.bbox())
        self.vertex_quad_tree = sub_layer

    def link_edges_together(self, edges, loop=False):
        """ Given a list of half edges, set their prev and next fields in order """
        #TODO: check this
        assert(all([isinstance(x, HalfEdge) for x in edges]))
        if loop:
            other = islice(cycle(edges), 1, None)
        else:
            other = edges[1:]
        for (a, b) in zip(edges, other):
            a.add_next(b)

    def force_all_edge_lengths(self, l):
        """ Force all edges to be of length <= l. If over, split into multiple lines
        of length l. """
        assert(l > 0)
        processed = set()
        all_edges = list(self.half_edges)
        while bool(all_edges):
            current = all_edges.pop(0)
            assert(current.index not in processed)
            if current.get_length_sq() > l:
                _, new_edge = current.split_by_ratio(r=0.5)
                if new_edge.get_length_sq() > l:
                    all_edges.append(new_edge)
                else:
                    processed.add(new_edge.index)

            if current.get_length_sq() > l:
                all_edges.append(current)
            else:
                processed.add(current.index)

    def constrain_to_circle(self, centre, radius, candidates=None, force=False):
        """ Limit all faces, edges, and vertices to be within a circle,
        adding boundary verts and edges as necessary """
        assert(isinstance(centre, np.ndarray))
        assert(isinstance(radius, float))
        assert(centre.shape == (2, ))

        #constrain faces
        faces = self.faces.copy()
        for f in faces:
            logging.debug("Constraining Face: {}".format(f))
            f.constrain_to_circle(centre, radius, candidates=candidates, force=force)

        #constrain free edges
        hedges = self.half_edges.copy()
        for he in hedges:
            logging.debug("Constraining Hedge: {}".format(he))
            if he.face is not None or he.marked_for_cleanup:
                continue
            he.constrain_to_circle(centre, radius, candidates=candidates, force=force)

        #constrain free vertices
        vertices = self.vertices.copy()
        for v in vertices:
            logging.debug("Constraining Vertex: {}".format(v))
            if not v.is_edgeless():
                continue
            if not v.within_circle(centre, radius):
                v.mark_for_cleanup()

    def constrain_to_bbox(self, bbox, candidates=None, force=False):
        """ Constrain the entire dcel to a bbox """
        assert(isinstance(bbox, np.ndarray))
        assert(bbox.shape == (4, ))

        faces = self.faces.copy()
        for f in faces:
            logging.debug("Constraining Face: {}".format(f))
            f.constrain_to_bbox(bbox, candidates=candidates, force=force)

        #constrain free edges
        hedges = self.half_edges.copy()
        for he in hedges:
            logging.debug("Constraining Hedge: {}".format(he))
            if he.face is not None or he.marked_for_cleanup:
                continue
            he.constrain_to_bbox(bbox, candidates=candidates, force=force)

        #constrain free vertices
        vertices = self.vertices.copy()
        for v in vertices:
            logging.debug("Constraining Vertex: {}".format(v))
            if not v.is_edgeless():
                continue
            if not v.within(bbox):
                v.mark_for_cleanup()

    #------------------------------
    # def Utilities
    #------------------------------

    def intersect_half_edges(self, edge_set=None):
        """ run a sweep line over the dcel,
        getting back halfedge intersections """

        li = LineIntersector(self)
        return li(edge_set=edge_set)

    def order_vertices(self, focus, vertices):
        """ Given a focus point and a list of vertices, sort them
            by the counter-clockwise angle position they take relative """
        assert(all([isinstance(x, Vertex) for x in vertices]))
        assert(isinstance(focus, np.ndarray))
        relative_positions = [v.loc - focus for v in vertices]
        zipped = zip(relative_positions, vertices)
        angled = [((degrees(atan2(loc[1], loc[0])) + 360) % 360, vert) for loc, vert in zipped]
        sorted_angled = sorted(angled)
        # rads = (np.arctan2(verts[:, 1], verts[:, 0]) + TWOPI) % TWOPI
        # ordered = sorted(zip(rads, opp_hedges))
        return [vert for loc, vert in sorted_angled]

    def create_corner_vertex(self, e1, e2, bbox):
        """ Given two intersections (0-3) describing the corner,
        create the vertex at the boundary of the bbox """
        assert(isinstance(e1, int))
        assert(isinstance(e2, int))
        assert(isinstance(bbox, np.ndarray))
        assert(len(bbox) == 4)

        if e1 == e2:
            raise Exception("Corner Edge Creation Error: edges are equal")
        if e1 % 2 == 0: #create the x vector
            v1 = np.array([bbox[e1], 0])
        else:
            v1 = np.array([0, bbox[e1]])
        if e2 % 2 == 0: #create the y vector
            v2 = np.array([bbox[e2], 0])
        else:
            v2 = np.array([0, bbox[e2]])
        #add together to get corner
        v3 = v1 + v2
        return self.new_vertex(*v3)

    def verify_all(self):
        """ Ensure all faces, edges and halfedges are coherent """
        reg_verts = set([x.index for x in self.vertices])
        reg_hedges = set([x.index for x in self.half_edges])
        reg_faces = set([x.index for x in self.faces])

        vert_hedges = set()
        for v in self.vertices:
            vert_hedges.update([x.index for x in v.half_edges])

        hedge_verts = set()
        hedge_nexts = set()
        hedge_prevs = set()
        hedge_faces = set()
        for h in self.half_edges:
            hedge_verts.add(h.origin.index)
            if h.next is not None:
                hedge_nexts.add(h.next.index)
            if h.prev is not None:
                hedge_prevs.add(h.prev.index)
            if h.face is not None:
                hedge_faces.add(h.face.index)

        face_edges = set()
        for f in self.faces:
            face_edges.update([x.index for x in f.edge_list])

        #differences:
        vert_hedge_diff = vert_hedges.difference(reg_hedges)
        hedge_vert_diff = hedge_verts.difference(reg_verts)
        hedge_nexts_diff = hedge_nexts.difference(reg_hedges)
        hedge_prevs_diff = hedge_prevs.difference(reg_hedges)
        hedge_faces_diff = hedge_faces.difference(reg_faces)
        face_edges_diff = face_edges.difference(reg_hedges)

        assert(all([not bool(x) for x in [vert_hedge_diff,
                                          hedge_vert_diff,
                                          hedge_nexts_diff,
                                          hedge_prevs_diff,
                                          hedge_faces_diff,
                                          face_edges_diff]]))




##-- imports
from __future__ import annotations

import warnings
import logging as logmod
import os
from itertools import cycle, chain
import time
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from functools import partial
from pathlib import Path
from typing import IO, List

##-- end imports

##-- logging
logging       = logmod.getLogger(__name__)
clingo_logger = logmod.getLogger(__name__ + ".ffi.clingo")
##-- end logging

import clingo
from clingo import Control, Function, Number, Symbol, parse_term
from jgdv._interfaces.solver import SolverWrapper_i
from jgdv.apis.clingo.ast import SolverAST
from jgdv.apis.clingo.solver import SolverModelResult, SolverWrapper_i

def clingo_intercept_logger(code, msg):
    """
    Intercepts messages from clingo, and controls
    the logging of them
    """
    msg = msg.replace("\n", "")
    match code:
        case clingo.MessageCode.AtomUndefined:
            clingo_logger.debug(msg)
        case clingo.MessageCode.RuntimeError:
            clingo_logger.exception(msg)
        case _:
            clingo_logger.info(msg)

def model_cb(self, model:clingo.Model):
    """
    Partial callback for clingo.
    Used for storing models.
    *Must* be constructed using functools.partial
    with 'self' bound to the SolverWrapper

    NOTE: Clingo Models are destroyed/reallocated on exit of the callback,
    Which is why we don't just store the model itself
    """
    self.results.append(SolverModelResult(model.symbols(atoms=True),
                                          model.symbols(shown=True),
                                          model.cost,
                                          model.number,
                                          model.optimality_proven,
                                          model.type))

@dataclass
class ClingoSolver(SolverWrapper_i):
    """
    An Oracle that uses Clingo as the solver
    """

    options      : list[str]    = field(kw_only=True, default_factory=list)
    ctl          : None|Control = field(init=False, default=None)
    program_name : str          = field(kw_only=True, default="base")

    def __post_init__(self):
        if not self.program and bool(self.input_files):
            warnings.warn("ClingoSolver created with no initial program or input files")

        self.options = self.options or ['-n', 1,
                                        '-c', f'horizon={2}']

        self.init_solver()

    def __repr__(self):
        return "[Clingo Solver: {}; Opts: {}; Models: {}]".format(self.ctl is not None,
                                                                  " ".join(self.options),
                                                                  len(self.results))
    def init_solver(self):
        self.results      = []
        default_grounding = [(self.program_name, [])]
        self.ctl          = Control([str(x) for x in self.options], logger=clingo_intercept_logger)

        for path in self.input_files:
            logging.debug("Clingo Loading: %s", path)
            assert(path.exists())
            self.ctl.load(str(path))

        try:
            if self.program:
                self.ctl.add(self.program_name, [], self.program)

        except Exception as err:
            logging.exception("Clingo Failed to add Program")
            logging.debug(self.program)
            raise err

        logging.debug("Initial Grounding of Program: %s", default_grounding)
        self.ctl.ground(default_grounding)
        logging.info("Clingo initialization complete")

    def solve(self, assignments:list[str|SolverAST|Symbol|tuple[bool, Symbol]]=None, fresh:bool=False, reground:list[tuple]=None) -> int:
        assignments = assignments or []

        if fresh or self.ctl is None:
            self.init_solver()

        if bool(reground):
            logging.debug("Grounding Program: %s", reground)
            self.ctl.cleanup()
            self.ctl.ground(reground)

        for sym in assignments:
            logging.debug("assigning: %s", sym)
            match sym:
                case SolverAST():
                    for truthy, sym in self.ast_to_clingo(sym):
                        self.ctl.assign_external(sym, truthy)
                case Symbol():
                    self.ctl.assign_external(sym, True)
                case bool(), Symbol():
                    self.ctl.assign_external(sym[1], sym[0])
                case str():
                    self.ctl.assign_external(parse_term(sym), True)
                case _:
                    raise Exception("Unrecognized situation fact")

        try:
            on_model_cb = partial(model_cb, self)
            logging.info("Running Program")
            self.ctl.solve(on_model=on_model_cb)
        except Exception as err:
            logging.warning("Clingo ran into a problem: %s", err)
            self.ctl = None

        logging.info("There are %s answer sets", len(self.results))
        return len(self.results)


    @property
    def metadata(self):
        return {
            "pid"            : os.getpid(),
            "source_files"   : [str(x) for x in self.input_files],
            "timestamp"      : self.timestamp,
            "mode"           : "multi_shot",
            "current_result" : self.current_answer,
            "result_size"    : len(self.results),
            "clingo_version" : clingo.__version__
        }


    def ast_to_clingo(self, *asts:TermAST) -> list[tuple[bool, Symbol]]:
        """
        Convert terms to clingo terms,
        not caring if they are #external or not
        """
        raise NotImplementedError()



""" Utilities for Cairo and DCEL """
##-- imports
from __future__ import annotations
import logging as root_logger
import pickle
from os.path import join, exists

##-- end imports

import pathlib as pl

logging = root_logger.getLogger(__name__)

def pickle_graph(node_dict, nx_graph, fpath:pl.Path):
    """
    Given a graph and dictionary of nodes,
    pickle it to the specified fpath
    """
    fpath.with_suffix(".graph.pickle".write_text(pickle.dumps(nx_graph))
    fpath.with_suffix(".nodes.pickle").write_text(pickle.dumps(node_dict))

def load_pickled_graph(fpath):
    """
    Given a fpath, load a pickled graph and
    its dictionary of nodes
    """
    node_dict = pickle.loads(fpath.with_suffix(".graph.pickle").read_text())
    nx_graph  = pickle.loads(fpath.with_suffix(".nodes.pickle").tead_text())
    return (node_dict, nx_graph)


def pickle_graph(node_dict, nx_graph, filename, directory="."):
    """ Given a graph and dictionary of nodes,
    pickle it to the specified filename """
    if not exists(directory):
        raise Exception("Directory not found: {}".format(directory))
    with open(join(directory, "{}_nxgraph.pickle".format(filename)), 'wb') as f:
        pickle.dump(nx_graph, f)
    with open(join(directory, "{}_node_dict.pickle".format(filename)), 'wb') as f:
        pickle.dump(node_dict, f)

def load_pickled_graph(filename, directory="."):
    """ Given a filename, load a pickled graph and
    its dictionary of nodes """
    if not (exists(directory) and \
             exists(join(directory, "{}_nxgraph.pickle".format(filename))) and \
             exists(join(directory, "{}_node_dict.pickle".format(filename)))):
        raise Exception("Missing file or directory for reconstruction")
    node_dict = None
    nx_graph = None
    with open(join(directory, "{}_nxgraph.pickle".format(filename)), 'rb') as f:
        nx_graph = pickle.load(f)
    with open(join(directory, "{}_node_dict.pickle".format(filename)), 'rb') as f:
        node_dict = pickle.load(f)
    return (node_dict, nx_graph)



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from collections import defaultdict
from jgdv.files.twitter.writers.util import parse_date

media_dict : Final[Callable] = lambda: defaultdict(list)

@dataclass
class TwitterTweet:
    id_s         : str
    base_user    : str
    is_quote     : bool                          = field(default=False)
    name         : str                           = field(init=False, default="Unknown")
    hash_tags    : List[str]                     = field(init=False, default_factory=list)
    quote        : Tuple[str, str, TwitterTweet] = field(init=False, default=None)
    reply_to     : Tuple[str, str]               = field(init=False, default=None)
    date         : datetime                      = field(init=False, default_factory=datetime.datetime.now)
    media        : dict                          = field(default_factory=media_dict)
    links        : list[str]                     = field(default_factory=list)
    level        : int                           = field(default=4)

    fav          : int = 0
    retweet      : int = 0
    text         : str = ""

    permalink_f  : str =  "[[https://twitter.com/{}/status/{}][/{}/{}]]"

    @staticmethod
    def build(base, tweet, users, tweet_lookup, level=4, is_quote=False) -> TwitterTweet:
        obj                  = TwitterTweet(tweet['id_str'], base, is_quote=is_quote, level=level)

        obj.name         = users.get(tweet.get('user', {}).get('id_str', None), {}).get('screen_name', None)
        obj.hash_tags    = [x.get('text', "") for x in tweet.get('entities', {}).get('hashtags', [])]
        obj.reply_to     = (tweet.get('in_reply_to_screen_name', None), tweet.get('in_reply_to_status_id_str', None))
        obj.fav          = str(tweet.get('favorite_count', 0))
        obj.retweet      = str(tweet.get('retweet_count', 0))
        obj.text         = tweet.get('full_text', "")
        date_str : None | str = tweet.get('created_at', None)
        if date_str:
            obj.date         = parse_date(date_str)

        urls             = tweet.get('entities', {}).get('urls', [])
        obj.links        = {x.get('expanded_url', None) for x in urls}

        try:
            quote_id         = tweet.get('quoted_status_id_str', None)
            quoted_tweet     = tweet_lookup.get(quote_id, {})
            quote_user_id    = quoted_tweet.get('user', {}).get('id_str', None)
            quoted_user_name = users.get(quote_user_id, {}).get('screen_name', "Unknown")
            quoted_tweet     = TwitterTweet.build(base, tweet_lookup[quote_id], users, tweet_lookup, level=level+1, is_quote=True)
            obj.quote        = (quoted_user_name, quote_id, quoted_tweet)
            # obj.media += quoted_tweet.media
            obj.links.update(quoted_tweet.links)
        except KeyError:
            pass

        media = TwitterTweet.get_tweet_media(tweet)
        obj.media = media
        return obj

    def __str__(self):
        output     = OrgStrBuilder()
        tags       = ""
        tag_offset = 0
        if bool(self.hash_tags):
            tags       =  ":{}:".format(":".join(self.hash_tags))
            tag_offset =  max(0, 80-len(self.at))
            tags       = (tag_offset * " ") + tags

        quote_header = ""
        if self.is_quote:
            quote_header = "Quote: "
        output.heading(self.level, quote_header, self.at, tags)

        with output.drawer("PROPERTIES") as dr:
            dr.add("PERMALINK", self.permalink(self.name, self.id_s))
            if self.reply_to is not None and self.reply_to[0] is not None:
                dr.add("REPLY_TO", self.permalink(*self.reply_to))
            if self.quote is not None:
                dr.add("QUOTE", self.permalink(*self.quote[:2]))
            dr.add("FAVOURITE_COUNT", self.fav)
            dr.add("RETWEET_COUNT", self.retweet)
            dr.add("DATE", self.date.strftime(OrgThreadWriter.date_re))
            if self.is_quote:
                dr.add("IS_QUOTE", "t")

        output.add(re.sub("\n\*", "\n-*", self.text))
        output.nl

        retarget= lambda x: OrgThreadWriter.retarget_url(self.base_user, x['url'])
        media_urls = []
        media_urls += [retarget(media) for media in self.media.get('photo', [])]
        media_urls += [retarget(media) for media in self.media.get('video', [])]

        if bool(media_urls):
            with output.drawer("MEDIA") as dr:
                dr.add_file_links(*media_urls)

        # Links
        if bool(self.links):
            with output.drawer("LINKS") as dr:
                dr.add_keyless(*self.links)

        if self.quote is not None:
            output.add(self.quote[2])

        return str(output)

    @property
    def at(self):
        return f"@{self.name}"

    @staticmethod
    def permalink(name, id_s):
        return TwitterTweet.permalink_f.format(name, id_s, name, id_s)

    @staticmethod
    def get_tweet_media(tweet) -> dict:
        media = media_dict()
        to_process = []
        to_process += tweet.get('entities', {}).get('media', [])
        to_process += tweet.get('extended_entities', {}).get('media', [])

        for ent in to_process:
            url        = ent.get('media_url_https', "")
            # filename   = pl.Path(url).name

            datum = {
                "url"          : url,
                "expanded_url" : ent.get("expanded_url", ""),
                "alt_text"     : ent.get('ext_alt_text', ""),
                "type"         : ent.get('type', "n/a"),
                }

            media[datum["type"]].append(datum)

            if datum['type'] == "video":
                video_datum = datum.copy()
                mp4s = [x for x in ent.get('video_info', {}).get('variants', []) if x.get('content_type', None) == "video/mp4"]
                best_variant = max(mp4s, key=lambda x: x.get('bitrate', 0))
                best_url = best_variant.get('url', "").split("?")[0]
                if bool(best_url):
                    video_datum['url'] = best_url
                    media['video'].append(video_datum)

        return media



#!/usr/bin/env python3
"""

"""
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
import warnings

import pytest

import tomlguard

logging = logmod.root

@pytest.mark.parametrize("ctor", [DootleReactorRunner])
class TestRunner:

    @pytest.fixture(scope="function")
    def setup(self, mocker):
        min_sleep   = 0.0
        config_dict = {"settings": {"tasks": {"sleep": {"task" : min_sleep, "subtask" : min_sleep, "batch": min_sleep}}}}
        doot.config = tomlguard.TomlGuard(config_dict)

    @pytest.fixture(scope="function")
    def cleanup(self):
        pass

    def test_initial(self, ctor, mocker, setup):
        ##-- setup
        tracker_m  = mocker.MagicMock(spec=TaskTracker_i)
        reporter_m = mocker.MagicMock(spec=ReportLine_i)
        runner     = ctor(tracker=tracker_m, reporter=reporter_m)
        ##-- end setup

        # Check:
        assert(isinstance(runner, TaskRunner_i))

    def test_tasks_execute(self, ctor, mocker, setup):
        ##-- setup
        tracker_m                       = mocker.MagicMock(spec=TaskTracker_i)
        reporter_m                      = mocker.MagicMock(spec=Reporter_i)
        runner                          = ctor(tracker=tracker_m, reporter=reporter_m)

        task1_m = mock_gen.mock_task(name="first", actions=0)
        task2_m = mock_gen.mock_task(name="second", actions=0)
        task3_m = mock_gen.mock_task(name="third", actions=0)
        tracker_m.__iter__.return_value  = [task1_m, task2_m, task3_m]

        expand_job  = mocker.spy(runner, "_expand_job")
        execute_task   = mocker.spy(runner, "_execute_task")
        execute_action = mocker.spy(runner, "_execute_action")
        ##-- end setup

        ##-- pre-check
        tracker_m.update_state.assert_not_called()
        ##-- end pre-check

        # Run
        runner()

        ##-- check result
        tracker_m.update_state.assert_called()
        assert(tracker_m.update_state.call_count == 3)
        for call in tracker_m.update_state.call_args_list:
            assert(call.args[0].name in ["first", "second", "third"])
            assert(call.args[1] is tracker_m.state_e.SUCCESS)

        expand_job.assert_not_called()
        execute_action.assert_not_called()

        execute_task.assert_called()
        assert(execute_task.call_count == 3)
        for call in execute_task.call_args_list:
            assert(call.args[0].name in ["first", "second", "third"])
        ##-- end check result

    def test_jobs_expand(self, ctor, mocker, setup):
        ##-- setup
        tracker_m                                     = mocker.MagicMock(spec=TaskTracker_i)
        reporter_m                                    = mocker.MagicMock(spec=Reporter_i)
        runner                                        = ctor(tracker=tracker_m, reporter=reporter_m)

        job1_m                                     = mock_gen.mock_job("first")
        job2_m                                     = mock_gen.mock_job("second")
        job3_m                                     = mock_gen.mock_job("third")

        tracker_m.__iter__.return_value               = [job1_m, job2_m, job3_m]

        expand_job                                 = mocker.spy(runner, "_expand_job")
        execute_task                                  = mocker.spy(runner, "_execute_task")
        execute_action                                = mocker.spy(runner, "_execute_action")
        ##-- end setup

        ##-- pre-check
        tracker_m.update_state.assert_not_called()
        ##-- end pre-check

        # Run
        runner()

        ##-- check
        tracker_m.update_state.assert_called()
        assert(tracker_m.update_state.call_count == 3)

        expand_job.assert_called()
        assert(expand_job.call_count == 3)

        execute_action.assert_not_called()
        execute_task.assert_not_called()
        ##-- end check

    def test_jobs_add_tasks(self, ctor, mocker, setup):
        ##-- setup
        tracker_m                          = mocker.MagicMock(spec=TaskTracker_i)
        reporter_m                         = mocker.MagicMock(spec=Reporter_i)
        runner                             = ctor(tracker=tracker_m, reporter=reporter_m)

        job1_m                          = mock_gen.mock_job("first")
        job2_m                          = mock_gen.mock_job("second")
        job3_m                          = mock_gen.mock_job("third")

        task_m                             = mock_gen.mock_task_spec("firstTask")

        job1_m.build.return_value       = [task_m]
        tracker_m.__iter__.return_value    = [job1_m, job2_m, job3_m]

        expand_job                      = mocker.spy(runner, "_expand_job")
        execute_task                       = mocker.spy(runner, "_execute_task")
        execute_action                     = mocker.spy(runner, "_execute_action")
        ##-- end setup

        ##-- pre-check
        tracker_m.update_state.assert_not_called()
        ##-- end pre-check

        ## Run:
        runner()

        ##-- Check
        tracker_m.add_task.assert_called()
        assert(tracker_m.add_task.call_count == 1)
        assert(tracker_m.add_task.call_args.args[0] is task_m)

        tracker_m.update_state.assert_called()
        assert(tracker_m.update_state.call_count == 3)

        expand_job.assert_called()
        assert(expand_job.call_count == 3)

        execute_action.assert_not_called()
        execute_task.assert_not_called()

    # @pytest.mark.xfail
    def test_tasks_execute_actions(self, ctor, mocker, setup):
        ##-- setup
        tracker_m                                  = mocker.MagicMock(spec=TaskTracker_i)
        reporter_m                                 = mocker.MagicMock(spec=Reporter_i)
        runner                                     = ctor(tracker=tracker_m, reporter=reporter_m)

        task1_m = mock_gen.mock_task("firstTask")
        task2_m = mock_gen.mock_task("secondTask")
        task3_m = mock_gen.mock_task("thirdTask")
        tracker_m.__iter__.return_value            = [task1_m, task2_m, task3_m]

        expand_job                              = mocker.spy(runner, "_expand_job")
        execute_task                               = mocker.spy(runner, "_execute_task")
        execute_action                             = mocker.spy(runner, "_execute_action")
        ##-- end setup

        ##-- pre-check
        ##-- end pre-check

        ## Run:
        runner()

        ##-- Check
        tracker_m.add_task.assert_not_called()
        expand_job.assert_not_called()

        tracker_m.update_state.assert_called()
        assert(tracker_m.update_state.call_count == 3)

        execute_task.assert_called()
        execute_action.assert_called()
        # TODO action_spec_m.__call__.assert_called()



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

NEIGHBOUR_CONDITION = lambda v, n: n.value.contains_vertex(v)

class LineSweep:

    def __call__(self, edge_set=None):
        #pylint: disable=unused-argument
        self.initialise_data(edge_set=edge_set)
        assert(bool(self.event_list))
        assert(bool(self.discovered))
        assert(not bool(self.results))
        assert(self.sweep_y == inf)

        #Main loop of the Intersection Algorithm
        while bool(self.event_list):
            logging.debug("--------------------")
            logging.debug("Event list remaining: {}".format(len(self.event_list)))
            logging.debug("\n".join([repr(x) for x in self.event_list]))

            curr_vert, curr_edge_list = self.get_next_event()
            logging.debug("Curr Vert: {}".format(curr_vert))
            self.update_sweep(curr_vert)
            self.debug_chain("Initial")

            logging.debug("Searching tree")
            closest_node, _ = self.search_tree(curr_vert)
            upper_set, contain_set, lower_set = self.determine_sets(curr_vert,
                                                                    closest_node,
                                                                    curr_edge_list.copy())
            self.report_intersections(curr_vert, upper_set, contain_set, lower_set)

            #todo: delete non-flat points of the non-flat event
            self.delete_values(contain_set, curr_x=curr_vert.loc[0])
            self.delete_values(lower_set, curr_x=curr_vert.loc[0])
            self.debug_chain("Deleted")

            #insert the segments with the status line a little lower
            candidate_lines = contain_set.union(upper_set)
            new_nodes = self.insert_values(candidate_lines, curr_x=curr_vert.loc[0])

            #Calculate additional events
            self.debug_chain("Inserted")

            self.handle_new_events(curr_vert, new_nodes)

        assert(self.sweep_y != inf)
        assert(not bool(self.event_list))
        return self.results

    def initialise_data(self, edge_set=None):
        """ Setup the internal structures to run a line sweep algorithm """
        logging.debug("Starting line intersection algorithm")
        self.results = []
        #setup the set of edges to intersect
        if edge_set is None:
            self.edge_set = self.dcel.half_edges.copy()
        else:
            assert(isinstance(edge_set, set))
            #get the twins as well
            twins = [x.twin for x in edge_set]
            edge_set.update(twins)
            self.edge_set = edge_set
        assert(self.edge_set is not None)
        self.lower_edges = [e for e in self.edge_set if not e.is_upper()]
        self.event_list = [HeapWrapper(x.origin, x, desc="initial") \
                           for x in self.edge_set.difference(self.lower_edges)]
        self.event_list += [HeapWrapper(x.origin, x.twin, desc="initial_twin") \
                            for x in self.lower_edges]
        heapq.heapify(self.event_list)
        self.discovered.update([x.ordinal for x in self.event_list])

        logging.debug("EdgeSet: {}, lower_edges: {}".format(len(self.edge_set),
                                                            len(self.lower_edges)))
        logging.debug("Event_list: {}".format(len(self.event_list)))

    def determine_sets(self, curr_vert, closest, candidates):
        """ TODO: describe """
        #pylint: disable=no-self-use
        assert(isinstance(curr_vert, Vertex))
        upper_set = set()
        contain_set = set()
        lower_set = set()

        if closest is not None:
            closest_segment = closest.value
            #if a line is horizontal, switch to a call(y) while,
            #and get lines until not in horizontal bounds
            candidate_nodes = closest.get_neighbours_while(partial(NEIGHBOUR_CONDITION, curr_vert))
            candidates += [x.value for x in candidate_nodes]
            candidates.append(closest_segment)

        logging.debug("Candidates: {}".format(candidates))
        #todo: adapt for horizontal lines
        #init a dictionary for intersections for each individual horizontal line
        #foreach individual horizontal line, test candidates until failure, then break
        for c in candidates:
            if c.origin == curr_vert:
                upper_set.add(c)
            elif c.twin.origin == curr_vert:
                lower_set.add(c)
            elif c.contains_vertex(curr_vert):
                contain_set.add(c)
        logging.debug("Upper Set: {}".format([x.index for x in upper_set]))
        logging.debug("Contain Set: {}".format([x.index for x in contain_set]))
        logging.debug("Lower Set: {}".format([x.index for x in lower_set]))
        return (upper_set, contain_set, lower_set)

    def handle_new_events(self, curr_vert, new_nodes):
        """ Given new nodes to add, process them """
        #pylint: disable=unused-argument
        logging.debug("Finding new Events")
        #todo: only find new events if there is a set of non-horizontal lines
        if not bool(new_nodes):
            closest_node, _ = self.search_tree(curr_vert)
            if closest_node is None:
                return
            left_n = closest_node.get_predecessor()
            right_n = closest_node.get_successor()
            if left_n is not None:
                self.find_new_events(left_n.value, closest_node.value, curr_vert.to_array())
            if right_n is not None:
                self.find_new_events(closest_node.value, right_n.value, curr_vert.to_array())

        else:
            logging.debug("New nodes added")
            assert(bool(self.status_tree))
            if not bool(new_nodes):
                return
            #TODO: this could be more efficient
            chain = self.status_tree.get_chain()
            ordinalered = [x for x in chain if x in new_nodes]
            logging.debug("New Nodes: {}".format([x.value.index for x in ordinalered]))
            leftmost = ordinalered[0]
            leftmost_n = leftmost.get_predecessor()
            if leftmost_n is not None and leftmost is not None:
                self.find_new_events(leftmost_n.value,
                                     leftmost.value,
                                     curr_vert.to_array())

            rightmost = ordinalered[-1]
            rightmost_n = rightmost.get_successor()

            if rightmost is not None and rightmost_n is not None:
                self.find_new_events(rightmost.value,
                                     rightmost_n.value,
                                     curr_vert.to_array())

    def find_new_events(self, a, b, loc):
        """ Given two edges and a location of the sweep line,
        find any valid intersection
        """
        assert(isinstance(loc, np.ndarray))
        assert(isinstance(a, HalfEdge))
        assert(isinstance(b, HalfEdge))

        logging.debug("Finding Events for: {}, {}, {}".format(a.index, b.index, loc))
        intersection = a.intersect(b)
        match_vert = None
        if intersection is None:
            logging.debug("No intersection")
            return
        if a.is_flat() or b.is_flat():
            logging.debug("Forcing same y position")
            intersection[1] = loc[1]

        #TODO: only works on cartesian
        if intersection[1] > loc[1]:
            logging.debug("Intersection too high: {} -- {}".format(intersection[1], loc[1]))
            return
        if intersection[1] < loc[1] or\
           (intersection[1] == loc[1] and loc[0] <= intersection[0]):
            logging.debug("Within bounds")
            match_vert = self.dcel.new_vertex(intersection)
            if match_vert in self.discovered:
                logging.debug("Vertex already discovered")
                return
        #finally, success!:
        if match_vert is not None:
            self.discovered.add(match_vert)
            wrapped = HeapWrapper(match_vert, a, desc="newEvent")
            wrapped2 = HeapWrapper(match_vert, b, desc="newEvent")
            logging.debug("Adding: {}".format(wrapped))
            heapq.heappush(self.event_list, wrapped)
            heapq.heappush(self.event_list, wrapped2)

    #------------------------------
    # def UTILITIES
    #------------------------------

    def debug_chain(self, the_str=None):
        """ log the flattened structure of the internal status tree """
        if the_str is None:
            the_str = ""
        chain = [x.value.index for x in self.status_tree.get_chain()]
        logging.debug("{} Tree Chain is: {}".format(the_str, chain))
        encountered = set()
        for x in chain:
            if x in encountered:
                raise Exception("Duplicated: {}".format(x))
            encountered.add(x)

    def report_intersections(self, v, u, c, l):
        """ Add intersection results when applicable """
        #todo: for each horizontal crossing point separately
        if sum([len(x) for x in [u, l, c]]) > 1:
            logging.debug("Reporting intersections")
            self.results.append(IntersectResult(v, u, c, l))

    def get_next_event(self):
        """ Get all segments that are on the current vertex """
        curr_vert, curr_edge_list = pop_while_same(self.event_list)
        assert(not bool(curr_edge_list) or all([isinstance(x, HalfEdge) for x in curr_edge_list]))
        return curr_vert, curr_edge_list

    def delete_values(self, values, curr_x):
        """ Remove a value from the status tree  """
        logging.debug("--------------------")
        logging.debug("Deleting values: {}".format([x.index for x in values]))
        assert(all([x.is_upper() for x in values]))
        chain = [x.value.index for x in self.status_tree.get_chain()]
        logging.debug("Chain: {}".format(chain))
        self.status_tree.delete_value(*values, cmp_data={'y':self.sweep_y, 'nudge': -SWEEP_NUDGE,
                                                         'x': curr_x - D_EPSILON})

        #Verify:
        chain = [x.value.index for x in self.status_tree.get_chain()]
        assert(all([x.index not in chain for x in values]))

    def insert_values(self, candidates, curr_x):
        """ Insert values into the status tree """
        logging.debug("Inserting values: {}".format([x.index for x in candidates]))
        flat_lines = set([x for x in candidates if x.is_flat()])
        #insert only non-flat lines
        to_add = candidates.difference(flat_lines)
        assert(all([x.is_upper() for x in to_add]))
        new_nodes = self.status_tree.insert(*to_add,
                                            cmp_data={'y':self.sweep_y, 'nudge': SWEEP_NUDGE,
                                                      'x': curr_x + D_EPSILON})
        new_nodes += self.status_tree.insert(*flat_lines,
                                             cmp_data={'y':self.sweep_y, 'nudge':SWEEP_NUDGE,
                                                       'x': curr_x + D_EPSILON})

        return new_nodes

    def update_sweep(self, curr):
        """ Increment the position of the frontier """
        assert(isinstance(curr, Vertex))
        candidate = curr.to_array()[1]
        if candidate > self.sweep_y:
            raise Exception("Sweep line moved in wrong direction: {} vs {}".format(candidate,
                                                                                   self.sweep_y))
        self.sweep_y = candidate

    def search_tree(self, curr):
        """ Search the status tree for a vertex """
        assert(isinstance(curr, Vertex))
        closest_node, d = self.status_tree.search(curr.to_array(),
                                                  cmp_data={'y': self.sweep_y,
                                                            'x': curr.to_array()[0]},
                                                  closest=True,
                                                  cmp_func=line_cmp_vert,
                                                  eq_func=line_eq_vert)

        if closest_node is not None:
            logging.debug("Closest Node: {}".format(closest_node.value.index))
        else:
            logging.debug("Closest Node is None")
        return closest_node, d



#!/usr/bin/env python2
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

printer = logmod.getLogger("doot._printer")

import inspect
import abc
import builtins
from typing import Type
import decorator

FUNC_WRAPPED     : Final[str]                = "__wrapped__"
jgdv_ANNOTATIONS : Final[str]                = "__JGDV_ANNOTATIONS__"
WRAPPER          : Final[str]                = "__wrapper"


class JGDVBaseDecorator(abc.ABC):
    """ Base Class for decorators that annotate action callables
      set self._annotations:dict to add annotations to fn.__DOOT_ANNOTATIONS (:set)
      implement self._wrapper to add a wrapper around the fn.
      TODO: set self._idempotent=True to only add a wrapper once
      """

    def __init__(self):
        self._idempotent  = False
        self._annotations =  set()

    def __call__(self, fn):
        if bool(self._annotations):
            self.annotate(fn, self._annotations)

        if not hasattr(self, WRAPPER):
            return fn

        if isinstance(fn, Type):
            return self.wrap_method(fn, fn.__call__, self._wrapper)

        decorated = decorator.decorate(fn, self._wrapper)
        return decorated

    @staticmethod
    def _strip_wrappers(fn:callable) -> callable:
        # if not hasattr(fn, FUNC_WRAPPED):
        #     return fn

        # return getattr(fn, FUNC_WRAPPED)
        return inspect.unwrap(fn)

    @staticmethod
    def has_annotations(fn, *keys) -> bool:
        base = JGDVDecorator._strip_wrappers(fn)
        if not hasattr(base, jgdv_ANNOTATIONS):
            return False

        annots = getattr(base, jgdv_ANNOTATIONS)
        return all(key in annots for key in keys)

    @staticmethod
    def annotate(fn:callable, annots:set) -> callable:
        base = JGDVDecorator._strip_wrappers(fn)
        if not hasattr(base, jgdv_ANNOTATIONS):
            setattr(base, jgdv_ANNOTATIONS, set())

        annotations = getattr(base, jgdv_ANNOTATIONS)
        for cls in getattr(fn, '__mro__', []):
            annotations.update(getattr(cls, jgdv_ANNOTATIONS, {}))

        annotations.update(annots)
        return fn

    @staticmethod
    def wrap_method(obj:Type, method:callable, wrapper:callable) -> Type:
        wrapped = decorator.decorate(method, wrapper)
        setattr(obj, method.__name__, wrapped)
        return obj

    @staticmethod
    def truncate_signature(fn):
        """
           actions are (self?, spec, state)
          with and extracted keys from the spec and state.
          This truncates the signature of the action to what is *called*, not what is *used*.

          TODO: could take a callable as the prototype to build the signature from
        """
        sig = inspect.signature(fn)
        min_index = len(sig.parameters) - len(getattr(fn, "_doot_keys"))
        newsig = sig.replace(parameters=list(sig.parameters.values())[:min_index])
        fn.__signature__ = newsig
        return fn




#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class JGDVDecorator_i:
    """ Base Class for decorators that annotate callables """

    def __init__(self, funcOrCls:Callable):
        ftz.update_wrapper(self, funcOrCls)
        self._func = func

    def __call__(self, *args, **kwargs):
        return self._func(*args, **kwargs)

class JGDVDelayDecorator_i:
    """ Base Class for decorators that take arguments, then later annotate callables

    https://stackoverflow.com/questions/9416947
      """

    def __init__(self):
        self._func = None

    def __call__(self, func):
        self._func = func
        ftz.update_wrapper(self._wrapper, self._func)
        return self._wrapper

    def _wrapper(self, *args, **kwargs):
        return



#!/usr/bin/env python3
"""
Colour Module, provides functions to convert between RGBA and HSLA
Code modifed from:
https://stackoverflow.com/questions/3018313/
"""
##-- imports
from __future__ import annotations

import types
import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from typing import TypeAlias
import numpy as np

__all__ = ["Colour", "rgba2hsla", "hsla2rgba"]

Colour : TypeAlias = np.array

def rgba2hsla(rgba:np.array):
    """
    Convert an RGBA colour to HSLA
    expects in range 0-1
    """
    c_max = np.max(rgba[:,:-1], axis=1)
    c_min = np.min(rgba[:,:-1], axis=1)
    delta = c_max - c_min

    lightness = (c_max + c_min) * 0.5
    sat = np.zeros((len(rgba),1))
    non_zero_deltas = delta != 0
    sat_div = (1 - np.abs(2 * lightness - 1))
    correct_for_zeros = ((sat_div == 0) * 1) + sat_div
    to_sat = non_zero_deltas * delta
    sat += (to_sat / correct_for_zeros).reshape((-1,1))

    hsla  = np.apply_along_axis(__to_hsla_row_func, 1, np.column_stack((c_max, delta, rgba, sat, lightness, rgba[:,-1])))
    return hsla

def hsla2rgba(hsla):
    """ Expects 0 <= h <= 360,
    0 <= sla <= 1
    """
    # from https://jsfiddle.net/Lamik/reuk63ay/91
    s = hsla[:,1]
    l = hsla[:,2]
    a = s * np.min((l,1-l), axis=0);

    offset = np.array([0, 8, 4])
    h_div = hsla[:,0] / 30

    ht_r = __to_rgba_calc_hue(h_div, 0, l, a)
    ht_g = __to_rgba_calc_hue(h_div, 8, l, a)
    ht_b = __to_rgba_calc_hue(h_div, 4, l, a)

    return np.column_stack((ht_r, ht_g, ht_b, hsla[:,-1]))

def __to_hsla_row_func(the_slice):
    c_max = the_slice[0]
    delta = the_slice[1]
    if delta == 0:
        delta = 1
    rgb = the_slice[2:6]
    hue = 0
    if c_max == rgb[0]:
        hue = np.mod((rgb[1] - rgb[2]) / delta, 6)
    elif c_max == rgb[1]:
        hue = ((rgb[2] - rgb[0]) / delta) + 2
    elif c_max == rgb[2]:
        hue = ((rgb[0] - rgb[1]) / delta) + 4

    hue *= 60
    return np.array((hue, *the_slice[-3:]))

def __to_rgba_calc_hue(h_div, n, l, a):
    k = ((h_div + n)) % 12
    ks_min = np.min((k-3, 9-k, np.ones(k.shape)), axis=0)
    ks_max = np.max((ks_min, -np.ones(k.shape)), axis=0)
    h = l - a * ks_max
    return h

def arrToHexColour(arr):
    """ Convert a triple of 0.0-1.0 values to hex """
    to255 = [int(255*x) for x in arr]
    toStrings = [format(x,'02x') for x in to255]
    return "#{}".format("".join(toStrings))



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from jgdv.files.org.base import OrgBuilderBase

@dataclass
class OrgDrawerBuilder(OrgBuilderBase):
    """
    A lazily build org drawer container,
    which aligns values in the block
    """

    owner               : OrgBuilderBase = field(default=None)
    name                : str           = field(default="")

    _contents     : List[Tuple[str, str]] = field(default_factory=list)
    _prop_pattern : str                   = ":{}:"
    _end          : str                   = ":END:"
    _max_key      : int                   = 0

    def add(self, *args):
        for name, contents in zip(args[::2], args[1::2]):
            self._contents.append((name, contents))
            self._max_key = max(len(name), self._max_key)

    def add_keyless(self, *args):
        for arg in args:
            self._contents.append(("", arg))

    def add_file_links(self, *args):
        as_links = [OrgDrawerBuilder.named_file_pattern.format(x, pl.Path(x).name) for x in args]
        self.add_keyless(*as_links)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        return None

    def __str__(self):
        """
        Write the drawer, while padding appropriately
        """
        output = []
        output.append(f":{self.name.upper()}:")
        for key, val in self._contents:
            if bool(key):
                key_f   = self._prop_pattern.format(key)
                pad_amt = 5 + max(0, (2 + self._max_key) - len(key_f))
                output.append(f"{key_f}{pad_amt*' '}{val}")
            else:
                output.append(val)

        output.append(self._end)
        output.append("")
        return "\n".join(map(str, output))



#!/usr/bin/env python3
"""

"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1

##-- end builtin imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from collections import ChainMap
from tomlguard.base import TomlTypes, GuardBase
from tomlguard.error import TomlAccessError
from tomlguard.utils.proxy import TomlGuardProxy, NullFallback

class TomlGuardIterProxy(TomlGuardProxy):
    pass

class TomlGuardIterFirstProxy(TomlGuardIterProxy):
    """
    A Proxy which handles lists by getting the first applicable value
    when passing through lists.
    So instead of:
    data.val.subval.listval[0].name
    it can be:
    data.first_of().val.subval.listval.name()

    """

    def __init__(self, data, types=None, index=None, subindex=None, fallback=NullFallback, nested=False):
        super().__init__(data, types=types, index=index, fallback=fallback)
        self._nested    = False
        self.__subindex = subindex or []

    def __repr__(self) -> str:
        type_str     = self._types_str()
        index_str    = ".".join(self._index())
        subindex_str = ".".join(self._subindex())
        return f"<TomlGuardIterFirstProxy: {index_str}:{subindex_str} ({self._fallback}) <{type_str}> >"

    def __call__(self, wrapper=None, fallback_wrapper=None) -> TomlTypes:
        if not self._nested and self._data is not NullFallback and all(isinstance(x, TomlGuardIterFirstProxy) for x in self._data):
            return self._get_first(wrapper=wrapper, fallback_wrapper=fallback_wrapper)

        return super().__call__(wrapper=wrapper, fallback_wrapper=fallback_wrapper)

    def _get_first(self, wrapper, fallback_wrapper) -> TomlTypes:
        """
        get the first value from any available table in an array
        """
        assert(isinstance(self._data, list))
        for val in self._data:
            assert(isinstance(val, TomlGuardIterFirstProxy))
            result = val(wrapper=wrapper, fallback_wrapper=fallback_wrapper)
            if result is None:
                continue
            return result

    def __iter__(self) -> Iterator[TomlTypes]:
        return iter(self())

    def _subindex(self, sub:str|None=None) -> list[str]:
        if sub is None:
            return self.__subindex[:]
        return self.__subindex[:] + [sub]

    def __getattr__(self, attr:str) -> TomlGuardProxy:
        try:
            match self._data:
                case list() if all(isinstance(x, (GuardBase, TomlGuardProxy)) for x in self._data):
                    return self._nested_inject(attr)
                case GuardBase():
                    return self._inject(self._data, attr=attr)
                case None:
                    raise TomlAccessError()
                case _:
                    return self._inject(attr=attr)
        except TomlAccessError:
            result = self._inject(clear=True, attr=attr)
            return result

    def _inject(self, val:tuple[Any]=NullFallback, attr:str|None=None, clear:bool=False) -> TomlGuardProxy:
        match val:
            case _ if clear:
                val = NullFallback
            case x if x is NullFallback:
                val = self._data
            case _:
                val = val[attr]

        return TomlGuardIterFirstProxy(val,
                                       types=self._types,
                                       index=self._index(attr),
                                       fallback=self._fallback)

    def _nested_inject(self, attr=None, clear=None) -> TomlGuardIterProxy|None:
        sub_proxies = []
        assert(isinstance(self._data, list))
        index    = self._index()
        subindex = self._subindex(attr)
        for data in self._data:
            match data:
                case x if x is NullFallback:
                    continue
                case TomlGuardIterFirstProxy() | GuardBase() if attr not in data:
                    continue
                case GuardBase():
                    new_proxy = TomlGuardIterFirstProxy(data[attr],
                                                        types=self._types,
                                                        fallback=self._fallback,
                                                        index=index,
                                                        subindex=subindex,
                                                        nested=True)
                case TomlGuardIterFirstProxy():
                    new_proxy = data[attr]

            sub_proxies.append(new_proxy)

        if not bool(sub_proxies):
            raise TomlAccessError()

        return TomlGuardIterFirstProxy(sub_proxies,
                                       types=self._types,
                                       fallback=self._fallback,
                                       index=index,
                                       subindex=subindex)

    def _match_type(self, val:TomlTypes) -> TomlTypes:
        return val

class TomlGuardIterAllProxy(TomlGuardIterProxy):
    """
    A Proxy for lists and dicts, which can flatten, or match particulars
    """

    def __init__(self, data, types=None, index=None, subindex=None, fallback=NullFallback, kind="first"):
        super().__init__(data, types=types, index=index, fallback=fallback)
        if self._fallback and not isinstance(self._fallback, Iterable):
            raise TypeError("Iter Proxy needs an iterable")
        self.__subindex = subindex or []

    def __repr__(self) -> str:
        type_str     = self._types_str()
        index_str    = ".".join(self._index())
        subindex_str = ".".join(self._subindex())
        return f"<TomlGuardIterAllProxy: {index_str}:{subindex_str} ({self._fallback}) <{type_str}> >"

    def __call__(self, wrapper=None) -> TomlTypes:
        self._notify()
        wrapper = wrapper or (lambda x: x)
        wrapped = wrapper(val)
        return self._match_type(wrapped)

    def __iter__(self) -> Iterator[TomlTypes]:
        return iter(self())

    def _subindex(self) -> list[str]:
        return self.__subindex[:]

    def _get_all(self) -> list[TomlTypes]:
        """
        Get all matching values from array of tables
        """
        match self._data, self._fallback:
            case [] | NullFallback(), NullFallback():
                pass
            case [] | NullFallback(), val:
                return val
            case [*vals], _:
                result = []
                for val in vals:
                    result += val
                return result

        base_index = ".".join(self._index())
        sub_index = ".".join(self._subindex())
        raise TomlAccessError(f"TomlGuardIterProxy Failure: {base_index}[?].{sub_index}")

    def _inject(self, val=NullFallback, attr=None, clear=None) -> TomlGuardIterProxy:
        match val:
            case NullFallback():
                val = self._data
            case _ if clear:
                val = NullFallback
            case _:
                val = val or self._data
        if clear:
            val = None

        return TomlGuardIterAllProxy(val,
                                     types=self._types,
                                     fallback=self._fallback,
                                     index=self._index(attr),
                                     subindex=new_index)

    def _match_type(self, val:TomlTypes) -> TomlTypes:
        return val

class TomlGuardIterFlatProxy(TomlGuardIterProxy):
    """
    A Proxy for lists and dicts, which can flatten, or match particulars
    """

    def __init__(self, data, types=None, index=None, subindex=None, fallback=None, kind="first"):
        super().__init__(data, types=types, index=index, fallback=fallback)
        if self._fallback and not isinstance(self._fallback, Iterable):
            raise TypeError("Iter Proxy needs an iterable")
        self.__subindex = subindex or []

    def __repr__(self) -> str:
        type_str     = self._types_str()
        index_str    = ".".join(self._index())
        subindex_str = ".".join(self._subindex())
        return f"<TomlGuardIterFlatProxy: {index_str}:{subindex_str} ({self._fallback}) <{type_str}> >"

    def __call__(self, wrapper=None) -> TomlTypes:
        self._notify()
        wrapper = wrapper or (lambda x: x)
        wrapped = wrapper(val)
        return self._match_type(wrapped)

    def __iter__(self) -> Iterator[TomlTypes]:
        return iter(self())

    def __getattr__(self, attr:str) -> TomlGuardIterProxy:
        try:
            match self._data:
                case GuardBase():
                    return self._inject(self._data[attr], attr=attr)
                case None:
                    raise TomlAccessError()
                case _:
                    return self._inject(attr=attr)
        except TomlAccessError:
            return self._inject(clear=True, attr=attr)

    def _subindex(self) -> list[str]:
        return self.__subindex[:]

    def _get_flat(self) -> GuardBase | list[TomlTypes]:
        match self._data:
            case [] | (None,):
                pass
            case [*vals]:
                return ChainMap(*(dict(x) for x in vals))

        match self._fallback:
            case (None,):
                pass
            case None:
                return None
            case dict() as x:
                return x

        base_index = ".".join(self._index())
        sub_index = ".".join(self._subindex())
        raise TomlAccessError(f"TomlGuardIterProxy Failure: {base_index}[?].{sub_index}")

    def _inject(self, val=None, attr=None, clear=None) -> TomlGuardIterProxy:
        new_index = self._subindex()
        if attr:
            new_index.append(attr)

        val = val or self._data
        if clear:
            val = None

        return TomlGuardIterFlatProxy(val, types=self._types, fallback=self._fallback, index=self._index(), subindex=new_index)

    def _match_type(self, val:TomlTypes) -> TomlTypes:
        return val

class TomlGuardIterMatchProxy(TomlGuardIterProxy):
    """
    A Proxy for lists and dicts, which can flatten, or match particulars
    """

    def __init__(self, data, types=None, index=None, subindex=None, fallback=None, kind="first"):
        super().__init__(data, types=types, index=index, fallback=fallback)
        if self._fallback and not isinstance(self._fallback, Iterable):
            raise TypeError("Iter Proxy needs an iterable")
        self.__subindex = subindex or []

    def __repr__(self) -> str:
        type_str     = self._types_str()
        index_str    = ".".join(self._index())
        subindex_str = ".".join(self._subindex())
        return f"<TomlGuardIterMatchProxy: {index_str}:{subindex_str} ({self._fallback}) <{type_str}> >"

    def __call__(self, wrapper=None) -> TomlTypes:
        self._notify()
        wrapper = wrapper or (lambda x: x)
        wrapped = wrapper(val)
        return self._match_type(wrapped)

    def __iter__(self) -> Iterator[TomlTypes]:
        return iter(self())

    def _subindex(self) -> list[str]:
        return self.__subindex[:]

    def _get_match(self) -> GuardBase | TomlTypes:
        """
        Get a table from an array if it matches a set of key=value pairs
        """
        for entry in self._data:
            try:
                for x in self._subindex():
                    entry = getattr(entry, x)
                if all(getattr(entry, x) == y for x,y in elf._kind.items()):
                    return entry
            except TomlAccessError:
                continue

        base_index = ".".join(self._index())
        sub_index = ".".join(self._subindex())
        raise TomlAccessError(f"TomlGuardIterProxy Match Failure: {base_index}[?].{sub_index} != {self._match}")

    def _inject(self, val=None, attr=None, clear=None) -> TomlGuardIterProxy:
        new_index = self._subindex()
        if attr:
            new_index.append(attr)

        val = val or self._data
        if clear:
            val = None

        return TomlGuardIterProxy(val, types=self._types, fallback=self._fallback, index=self._index(), subindex=new_index)

    def _match_type(self, val:TomlTypes) -> TomlTypes:
        return val
